import logging
import warnings
from typing import Any, Literal, Optional, Tuple, Union

import numpy as np
from numpy.typing import ArrayLike, NDArray
from scipy.ndimage import zoom
from scipy.signal import windows


def check_sr_format(sr: Union[int, float]) -> int:
    try:
        sr = int(sr)
    except Exception as e:
        raise TypeError(f"Sample rate not transformable to integer: {e}")
    if sr <= 0:
        raise ValueError("Sample rate must be greater than zero.")
    return sr


def check_signal_format(data: ArrayLike) -> NDArray[np.float32]:
    data = np.asarray(data, dtype=np.float32)
    if data.ndim != 1:
        raise ValueError("Signal must be a 1D array.")
    if not np.issubdtype(data.dtype, np.floating):
        raise TypeError("Signal must be an array of type float.")
    # if np.max(data) > 1:
        # scale to -1 to 1
        # TODO
    return data


def exclude_trailing_and_leading_zeros(envelope: ArrayLike) -> NDArray[np.float32]:
    """
    Removes leading and trailing zeros from a NumPy array.

    This function identifies and excludes the leading and trailing
    zeros in the input array `envelope`. It finds the first and last non-zero
    elements, and returns a new array that starts from the first non-zero value
    and ends at the last non-zero value.

    Args:
        envelope : ArrayLike
            A 1D NumPy array containing numerical values,
            potentially with leading and/or trailing zeros.

    Returns:
        np.ndarray
            A 1D NumPy array with leading and trailing zeros excluded.
            The returned array will only contain values between the first
            and last non-zero elements from the original array.

    Example:
        >>> import numpy as np
        >>> arr = np.array([0, 0, 1, 2, 3, 0, 0])
        >>> exclude_trailing_and_leading_zeros(arr)
        array([1, 2, 3])

    Notes:
        - If the array consists entirely of zeros, an empty array will be returned.
        - The function assumes that the input array is 1D.

    """
    if envelope.ndim != 1:
        raise ValueError("Input array must be 1D.")

    if np.all(envelope == 0):  # check if all zeros and return empty array
        return np.array([])

    try:
        # Exclude trailing zeros (identify last non-zero element)
        non_zero_end = np.argmax(envelope[::-1] > 0)
        envelope = envelope[:len(envelope) - non_zero_end]

        # Exclude leading zeros (identify first non-zero value)
        non_zero_start = np.argmax(envelope > 0)
        envelope = envelope[non_zero_start:]

    except Exception as e:
        logging.error(f"An error occurred: {e}")
        raise

    return envelope


def probability_mass_function(envelope: NDArray[np.float32]) -> NDArray[np.float32]:
    return envelope / np.sum(envelope)


def cumulative_distribution_function(envelope: NDArray[np.float32]) -> NDArray[np.float32]:
    return np.cumsum(probability_mass_function(envelope))


def extract_all_features(
    data: ArrayLike,
    sr: int,
    kernel_size: Optional[int] = None,
    n_dominant_freqs: int = 1,
    plot: bool = False,
    plot_kwargs: dict[str, Any] = {},
    spec_kwargs: dict[str, Any] = {},
    **kwargs: dict[str, Any]
) -> dict[str, Any]:
    """
    Extracts a comprehensive set of temporal and spectral features from a signal.

    Args:
        data (ArrayLike): Input 1D signal.
        sr (int): Sampling rate in Hz.
        kernel_size (Optional[int]): Size of smoothing kernel for amplitude envelope.
        n_dominant_freqs (int): Number of dominant frequencies to extract per frame.
        **kwargs (dict[str, Any]): Optional parameters for dominant frequency estimation.

    Returns:
        dict: Dictionary of extracted features.
    """
    from .spectral import spectral_features
    from .spectrotemporal import spectrotemporal_features
    from .temporal import temporal_features

    data = check_signal_format(data)
    sr = check_sr_format(sr)

    temporal_feats = temporal_features(data, sr, kernel_size)
    spectral_feats = spectral_features(data, sr)
    spectrotemporal_feats = spectrotemporal_features(data, sr, n_dominant_freqs, **kwargs)

    if plot:
        from biosonic.plot import plot_features
        plot_features(data, sr, {**temporal_feats, **spectral_feats, **spectrotemporal_feats}, spec_kwargs, **plot_kwargs)
    return {**temporal_feats, **spectral_feats, **spectrotemporal_feats}


def transform_spectrogram_for_nn(
        data: ArrayLike,
        sr: Optional[int] = None,
        values_type: str = 'float32',
        add_channel: bool = True,
        data_format: Literal['channels_last', 'channels_first'] = 'channels_first',
        f_min: Optional[float] = None,
        f_max: Optional[float] = None,
        resize: Optional[Tuple[int, int]] = None,
        **kwargs: Any,
    ) -> ArrayLike:
    """
    Prepares a spectrogram for input into a neural network by normalizing, casting type,
    and optionally adding a channel dimension.

    Parameters:
        data : Union[np.ndarray, Tuple[np.ndarray, np.ndarray, np.ndarray]],
            Input signal. Either as precomputed spectrogram (S, t, f) or 1D signal array.
        sr : Optional[int]
            Sampling rate in Hz. Needed when passing signal as 1D array.
        values_type : str
            Data type to cast the spectrogram to (e.g., 'float32', 'float64'). Defaults to 'float32'
        add_channel : bool
            Whether to add a greyscale channel dimension.
        data_format : Literal['channels_last', 'channels_first']
            Specifies channel dimension placement when `add_channel` is True.
            - 'channels_last' results in shape (H, W, 1) - e.g. for TensorFlow/Keras
            - 'channels_first' results in shape (1, H, W) - e.g. for PyTorch
        fmin : Optional[float]
            Lower frequency bound in Hz.
        fmax : Optional[float]
            Upper frequency bound in Hz.
        resize : Optional[Tuple[int, int]]
            Resize output to (height, width).
        **kwargs : dict
            Additional keyword arguments for spectrogram calculation.
    Returns:
        ArrayLike :
            The transformed spectrogram, normalized to [0, 1], cast to the specified
            data type, and optionally with a channel dimension added.

    Example:
        >>> import numpy as np
        >>> spec = np.random.rand(128, 128) * 255  # Example spectrogram
        >>> processed = transform_spectrogram_for_nn(spec, values_type='float32',
        ...                                          add_channel=True, data_format='channels_last')
        >>> processed.shape
        (128, 128, 1)
        >>> processed.dtype
        dtype('float32')
    """
    from .spectrotemporal import spectrogram

    # precomputed spectrogram
    if isinstance(data, tuple) and len(data) == 3:
        spec, t, f = data

    # raw signal + sr
    elif isinstance(data, np.ndarray):
        if sr is None:
            raise ValueError("sr must be provided when passing a signal array.")
        spec, t, f = spectrogram(
            data=data,
            sr=sr,
            complex_output=False,
            **kwargs
        )

    if f_min is not None or f_max is not None:
        freq_mask = np.ones_like(f, dtype=bool)
        if f_min is not None:
            freq_mask &= f >= f_min
        if f_max is not None:
            freq_mask &= f <= f_max
        spec = spec[freq_mask, :]
        f = f[freq_mask]

    if spec.max() - spec.min() == 0:
        warnings.warn(f"Spectrogram contains no information (values in range [{spec.min()}, {spec.max()}]).", RuntimeWarning)
    else:
        # min-max-scale to range [0, 1]
        spec = (spec - spec.min()) / (spec.max() - spec.min())

    if resize is not None:
        target_height, target_width = resize
        zoom_factors = (target_height / spec.shape[0], target_width / spec.shape[1])
        spec = zoom(spec, zoom_factors, order=1)  # bilinear interpolation

    # add channel dimension
    if add_channel:
        if data_format == 'channels_last':
            spec = np.expand_dims(spec, axis=-1)  # (H, W, 1)
        else:
            spec = np.expand_dims(spec, axis=0)   # (1, H, W)

    # convert to desired bit depth
    return spec.astype(values_type)


def shannon_entropy(
        prob_dist: ArrayLike,
        unit: Literal["bits", "nat", "dits", "bans", "hartleys"] = "bits",
        norm: bool = True
    ) -> Tuple[float, float]:
    """
    Calculate the Shannon entropy of a probability distribution.

    Parameters
    ----------
    prob_dist : ArrayLike
        A 1D array-like object representing a probability distribution.
        Values should be non-negative and typically sum to 1.

    unit : {'bits', 'nat', 'dits', 'bans', 'hartleys'}, default='bits'
        The desired output unit, determining the logarithmic base to use for entropy calculation:
        - 'bits': base 2 (Shannon entropy)
        - 'nat' : base e (natural logarithm)
        - 'dits', 'bans', 'hartleys': base 10

    norm : bool, default=True
        Whether to normalize the entropy by dividing by the maximum possible entropy
        for the given distribution length, resulting in a value in [0, 1].

    Returns
    -------
    entropy : float
        The Shannon entropy of the distribution.

    max_entropy : float
        The maximum possible entropy for the distribution length (either normalized to 1 or
        log-base(len(dist)) if not normalized).

    Raises
    ------
    ValueError
        If an invalid unit is specified.

    Examples
    --------
    >>> shannon_entropy([0.5, 0.5])
    (0.0, 1.0)

    >>> shannon_entropy([0.9, 0.1], unit="nat", norm=False)
    (0.3250829733914482, 0.6931471805599453)
    """
    if unit == "bits":
        log_: Any = np.log2
    elif unit == "nat":
        log_ = np.log
    elif unit in ["dits", "bans", "hartleys"]:
        log_ = np.log10
    else:
        raise ValueError(f'Invalid unit for power spectral entropy: {unit} Must be in ["bits", "nat", "dits", "bans", "hartleys"]')

    if np.all(prob_dist == prob_dist[0]):
        return 0.0, 1.0 if norm else log_(len(prob_dist))

    if norm:
        H = np.negative(np.sum(prob_dist * (log_(prob_dist)/log_(len(prob_dist)))))
        max = 1
    else:
        H = np.negative(np.sum(prob_dist * log_(prob_dist)))
        max = log_(len(prob_dist))

    return float(H), float(max)


def hz_to_mel(
        f: Union[ArrayLike, float],
        a: Optional[float] = None,
        b: Optional[float] = None,
        corner_frequency: Optional[float] = None,
        after: Literal["fant", "koenig", "oshaughnessy", "umesh"] = "oshaughnessy"
    ) -> Union[ArrayLike, float]:
    """
    Converts a frequency or array of frequencies in Hertz to the Mel scale
    using one of several proposed formulas.

    Parameters:
        f : Union[ArrayLike, np.floating])
            Frequency or array of frequencies in Hz.
        a : Optional[float])
            Scaling factor for the selected formula. Default depends on `after`.
        b : Optional[float]
            Denominator or offset parameter. Default depends on `after`.
        corner_frequency : Optional[float]
            If corner frequency other than 1000 Hz is desired, given in Hz. If provided, calculations are based on Fant (1970).
        after (Literal): Choice of Mel scale formula to use.
            - 'fant': Classic formula with `a=b=1000`: `F_m = a * np.log(1 + f / b)`
            - 'koenig': (Not yet implemented)
            - 'oshaughnessy' or 'beranek': Commonly used formula as fant, but with `a=2595`, `b=700`
            - 'umesh': Formula using rational function with `a=0.0004`, `b=0.603`.

    Returns:
        np.ndarray: Frequencies mapped to the Mel scale.


    References:
        1. Fant G. 1970 Acoustic Theory of Speech Production. The Hague: Mouton & Co.
        2. Umesh S, Cohen L, Nelson D. 1999 Fitting the Mel scale. In 1999 IEEE International
           Conference on Acoustics, Speech, and Signal Processing. Proceedings. ICASSP99 (Cat. No.99CH36258),
           pp. 217–220 vol.1. Phoenix, AZ, USA: IEEE. (doi:10.1109/ICASSP.1999.758101)
        3. D. O'Shaughnessy, ”Speech Communication - Human and Machine” Addison- Wesley, New York, 1987. As cited in [2]

    """
    f_arr = np.asarray(f, dtype=np.float32)

    if corner_frequency is not None:
        a, b = corner_frequency, corner_frequency
        after = "fant"

    if after == "fant":
        a = 1000.0 if a is None else a
        b = 1000.0 if b is None else b
        F_m = a * np.log(1 + f_arr / b)
    elif after in ["oshaughnessy", "beranek"]:
        a = 2595.0 if a is None else a
        b = 700.0 if b is None else b
        F_m = a * np.log(1 + f_arr / b)
    elif after == "umesh":
        a = 0.0004 if a is None else a
        b = 0.603 if b is None else b
        F_m = f_arr / (a * f_arr + b)
    elif after == "koenig":
        raise NotImplementedError("The 'koenig' Mel scale formula is not yet implemented.")
    else:
        raise ValueError(f"Unknown Mel scale method: '{after}'")

    return F_m


def mel_to_hz(
    m: Union[ArrayLike, float],
    a: Optional[float] = None,
    b: Optional[float] = None,
    corner_frequency: Optional[float] = None,
    after: Literal["fant", "koenig", "oshaughnessy", "umesh"] = "oshaughnessy"
) -> Union[ArrayLike, float]:
    # TODO Slaney
    """
    Converts a Mel scale value or array of values to frequency in Hertz,
    using one of several inverse Mel formulas.

    Parameters
    ----------
    m : Union[ArrayLike, float]
        Mel frequency or array of Mel frequencies.
    a : Optional[float]
        Scaling factor for the selected formula. Default depends on `after`.
    b : Optional[float]
        Denominator or offset parameter. Default depends on `after`.
    corner_frequency : Optional[float]
        If a corner frequency other than 1000 Hz is desired, given in Hz. If provided, calculations are based on Fant (1970).
    after : Literal
        Choice of Mel scale formula to use.
        - 'fant': Classic formula: `f = b * (exp(m/a) - 1)` with `a=b=1000`
        - 'oshaughnessy' or 'beranek': Common formula: `a=2595`, `b=700`
        - 'umesh': Rational model: `f = b * m / (1 - a * m)`
        - 'koenig': (Not yet implemented)

    Returns
    -------
    Union[ArrayLike, float]
        Frequency or array of frequencies in Hertz.

    References
    ----------
    1. Fant G. 1970 Acoustic Theory of Speech Production.
    2. Umesh S, Cohen L, Nelson D. 1999 Fitting the Mel scale. ICASSP.
    3. D. O'Shaughnessy, "Speech Communication - Human and Machine", 1987.
    """
    m_arr = np.asarray(m, dtype=np.float32)

    if corner_frequency is not None:
        a = corner_frequency
        b = corner_frequency
        after = "fant"

    if after == "fant":
        a = 1000.0 if a is None else a
        b = 1000.0 if b is None else b
        f_hz = b * (np.exp(m_arr / a) - 1)
    elif after in ["oshaughnessy", "beranek"]:
        a = 2595.0 if a is None else a
        b = 700.0 if b is None else b
        f_hz = b * (np.exp(m_arr / a) - 1)
    elif after == "umesh":
        a = 0.0004 if a is None else a
        b = 0.603 if b is None else b
        denominator = 1 - a * m_arr
        if np.any(denominator == 0):
            raise ZeroDivisionError("Denominator in Umesh model became zero.")
        f_hz = b * m_arr / denominator
    elif after == "koenig":
        raise NotImplementedError("The 'koenig' Mel scale formula is not yet implemented.")
    else:
        raise ValueError(f"Unknown Mel scale method: '{after}'")

    return f_hz


def frame_signal(
        data: ArrayLike,
        sr: int,
        window_length: int = 512,
        timestep: float = 0.01,
        normalize: bool = False
    ) -> ArrayLike:
    samples_step = timestep * sr
    data = np.pad(data, int(window_length / 2), mode='reflect')

    frame_num = int((len(data) - window_length) / samples_step) + 1
    frames = np.zeros((frame_num, window_length))

    for n in range(frame_num):
        start = int(n * samples_step)
        frames[n] = data[start:start + window_length]

    if normalize:
        frames = [frame / np.mean(frame) if frame.any() else frame for frame in frames]  # skip normalization for frames with all 0

    return frames


def window_signal(
        data: ArrayLike,
        sr: int,
        window_length: int = 512,
        window: Union[str, ArrayLike] = "hann",
        timestep: float = 0.01,
        normalize: bool = False
) -> ArrayLike:

    if isinstance(window, str):
        try:
            window = windows.get_window(window, window_length)
        except ValueError as e:
            raise ValueError(f"Invalid window type: {window}") from e
    else:
        window = np.asarray(window)
        if not isinstance(window, np.ndarray):
            raise TypeError("'window' must be either a string or a 1D NumPy array.")

    frames = frame_signal(data, sr, window_length, timestep, normalize)

    return frames * window



import warnings
from typing import Any, Dict, Literal, Optional, Tuple, Union

import numpy as np
from numpy.typing import ArrayLike, NDArray
from scipy import signal
from scipy.stats import kurtosis, skew

from .utils import (
    check_signal_format,
    check_sr_format,
    cumulative_distribution_function,
    exclude_trailing_and_leading_zeros,
    shannon_entropy
)


def amplitude_envelope(
        data: ArrayLike,
        kernel_size: Optional[int] = None,
        remove_trailing_zeros: bool = True,
        avoid_zero_values: bool = False
    ) -> NDArray[np.float64]:
    """
    Computes the amplitude envelope of a signal using the Hilbert transform.

    The function applies the Hilbert transform to calculate the analytic signal of the input data
    and then computes the envelope as the absolute value of the analytic signal. Optionally,
    a smoothing kernel (Daniell kernel, moving average) can be applied to smooth the envelope.

    Args:
        data : ArrayLike
            A 1D array-like representing the input signal.
        kernel_size : Optional[int]
            The size of the moving average kernel to smooth the envelope. Must be an uneven number.
            If `None`, no smoothing is applied (default is `None`).
        remove_trailing_zeros : Optional[bool]
            If `True`, removes leading and trailing zeros of the envelope after applying Hilbert transform.
            Defaults to `True`
        avoid_zero_values : Optional[bool]
            If `True`, replaces zero values with 1e-10 after applying Hilbert transform and removing leading and trailing zeros.
            Defaults to `False`

    Returns:
        NDArray[np.float64]
            A 1D NumPy array of float64 values representing the amplitude envelope of the signal.

    Example:
        >>> import numpy as np
        >>> signal = np.array([0, 1, 2, 3, 2, 1, 0], dtype=np.float64)
        >>> envelope = amplitude_envelope(signal, kernel_size=3)
        >>> print(envelope)
        [0.  1.  2.  3.  2.  1.  0.]

    Notes:
        - The Hilbert transform is used to generate the analytic signal.
        - If `kernel_size` is provided, a moving average filter (Daniell kernel) is applied to smooth the envelope.
        - The envelope is processed to remove any leading or trailing zeros using the `exclude_trailing_and_leading_zeros` function.
    """
    if len(data) == 0:
        warnings.warn("Input signal is empty; returning an empty array.", RuntimeWarning)
        return np.array([], dtype=np.float64)

    data = check_signal_format(data)

    # compute the analytic signal using Hilbert transform
    analytic_signal = signal.hilbert(data)
    envelope = np.abs(analytic_signal)

    if kernel_size is not None:
        # apply Daniell kernel (moving average kernel) only if kernel_size is provided
        kernel = np.ones(kernel_size) / kernel_size

        # apply kernel to smooth envelope
        envelope = signal.convolve(envelope, kernel, mode='same')

    if remove_trailing_zeros:
        envelope = exclude_trailing_and_leading_zeros(envelope)

    if avoid_zero_values:
        # replace values close to zero to avoid log conflicts in following calculations
        envelope = np.where(envelope < 1e-10, 1e-10, envelope)

    return envelope


def duration(data: ArrayLike, sr: int, exclude_surrounding_silences: Optional[bool] = True) -> float:
    """
    Returns the duration in seconds of a signal.

    Args:
        data : ArrayLike
            The audio signal data.
        sr : int
            The sample rate of the audio signal (samples per second).
        exclude_surrounding_silences : bool, optional
            If True, leading and trailing silences (zeros) in the audio
            signal will be removed before calculating duration.
            Defaults to False.

    Returns:
        float
            Duration of the (possibly trimmed) audio signal in seconds.
    """
    data = check_signal_format(data)
    check_sr_format(sr)

    if exclude_surrounding_silences:
        data = exclude_trailing_and_leading_zeros(data)

    return len(data) / sr


def temporal_quartiles(data: ArrayLike, sr: int, kernel_size: Optional[int] = None) -> Tuple[float, float, float]:
    """
    Computes the temporal quartiles (Q1, median, Q3) of the amplitude envelope.

    This function takes a signal and its sample rate, then normalizes the
    signal to obtain the amplitude envelope. It calculates the cumulative sum of the normalized
    envelope and uses the `searchsorted` function to find the time points corresponding to the
    25th, 50th, and 75th percentiles of the cumulative amplitude. These correspond to the temporal
    quartiles (Q1, median, Q3) of the signal.

    Args:
        data : ArrayLike
            The signal, stored as a 1D NumPy array of float64 values.
        sr : int
            The sample rate of the signal (Hz).
        kernel_size : Optional[int]
            Optional smoothing kernel applied to the amplitude envelope.

    Returns:
        Tuple[float, float, float]
            A tuple containing the temporal quartiles (Q1, median, Q3),
            in seconds.

    Example:
        >>> signal = np.array([0, 1, 2, 3, 2, 1, 0], dtype=np.float64)
        >>> sr = 44100  # Sample rate of 44.1 kHz
        >>> temporal_quartiles(signal, sr)
        (0.03, 0.05, 0.07)

    Notes:
        - The `searchsorted` function is used to find the indices corresponding to the quartiles
          in the cumulative amplitude envelope. The result is scaled by the sample rate to return
          time values in seconds.
    """
    data = check_signal_format(data)
    check_sr_format(sr)

    if len(data) == 0:
        raise ValueError("Input is empty")
    if np.all(data == 0):
        raise ValueError("Signal contains no nonzero values")

    envelope = amplitude_envelope(data, kernel_size)

    cdf = cumulative_distribution_function(envelope)

    # temporal quartiles (Q1, median, Q3)
    t_q1 = float(np.searchsorted(cdf, 0.25) / sr)
    t_median = float(np.searchsorted(cdf, 0.5) / sr)
    t_q3 = float(np.searchsorted(cdf, 0.75) / sr)

    return (t_q1, t_median, t_q3)


def temporal_sd(data: ArrayLike, sr: int, kernel_size: Optional[int] = None) -> float:
    """
    Computes the temporal standard deviation of the amplitude envelope.

    Args:
        data : ArrayLike
            The signal, stored as a 1D NumPy array of float64 values.
        sr : int
            The sample rate of the signal (Hz).
        kernel_size : Optional[int]
            Optional smoothing kernel applied to the amplitude envelope.

    Returns:
        float
            The standard deviation of the amplitude envelope
    """
    data = check_signal_format(data)
    check_sr_format(sr)

    return float(np.std(amplitude_envelope(data, kernel_size)))


def skewness(data: ArrayLike, sr: int, kernel_size: Optional[int] = None) -> Optional[float]:
    """
    Computes the temporal skew of the signal.

    Args:
        data : ArrayLike
            The signal, stored as a 1D NumPy array of float64 values.
        sr : int
            The sample rate of the signal (Hz).
        kernel_size : Optional[int]
            Optional smoothing kernel applied to the amplitude envelope.

    Returns:
        float
            The temporal skew
    """
    data = check_signal_format(data)
    check_sr_format(sr)

    skew_ = skew(amplitude_envelope(data, kernel_size))
    if skew_ is None:
        warnings.warn("All values are equal, returning None for skew.")

    return float(skew_)


def temporal_kurtosis(data: ArrayLike, sr: int, kernel_size: Optional[int] = None) -> Optional[float]:
    """
    Computes the temporal kurtosis of the signal.

    Args:
        data : ArrayLike
            The signal, stored as a 1D NumPy array of float64 values.
        sr : int
            The sample rate of the signal (Hz).
        kernel_size : Optional[int]
            Optional smoothing kernel applied to the amplitude envelope.

    Returns:
        float
            The temporal kurtosis
    """
    data = check_signal_format(data)
    sr = check_sr_format(sr)

    kurtosis_ = kurtosis(amplitude_envelope(data, kernel_size))

    if kurtosis_ is None:
        warnings.warn("All values are equal, returning None for kurtosis.")

    return float(kurtosis_)


def temporal_entropy(
        data: ArrayLike,
        unit: Literal["bits", "nat", "dits", "bans", "hartleys"] = "bits",
        *args: Any,
        **kwargs: Any
    ) -> Tuple[float, float]:
    """
    Calculates the entropy of the amplitude envelope as follows:
    1. Compute the amplitude envelope.
    2. Calculate a discrete probability distribution from the histogram of the envelope.
    3. Calculate Shannon-Wiener entropy of the probability distribution.

    Args:
        data : ArrayLike
            Input signal as a 1D ArrayLike.
        unit : str, optional
            Desired unit of the entropy, determines the logarithmic base used for calculatein.
            Choose from "bits" (log2), "nat" (ln), or "dits"/"bans"/"hartleys" (log10).
            Defaults to "bits".

    Returns:
        float
            Temporal entropy.
    References:
        1. https://de.mathworks.com/help/signal/ref/spectralentropy.html accessed January 13th, 2025. 18:34 pm
        2. https://docs.scipy.org/doc/scipy-1.15.2/reference/generated/scipy.stats.entropy.html accessed May 20th 2025, 11:32 am
        3. Shannon C. E. 1948 A mathematical theory of communication. The Bell System Technical Journal XXVII.

    Notes:
        Rounds
    """
    envelope = amplitude_envelope(data)
    hist, _ = np.histogram(envelope, bins="auto", density=False)
    hist = hist[hist > 0]
    norm_counts = hist / np.sum(hist)

    return shannon_entropy(norm_counts, unit, *args, **kwargs)


def temporal_features(
        data: NDArray[np.float32],
        sr: int,
        kernel_size: Optional[int] = None,
        ) -> Dict[str, Union[float, NDArray[np.float32]]]:
    """
    Extracts a set of temporal features from the amplitude envelope of a signal.

    Returns
    -------
        dict
        {
        "t_q1": float,
        "t_median": float,
        "t_q3": float,
        "temporal_sd": float,
        "temporal_skew": float,
        "temporal_kurtosis": float,
        "amplitude_envelope": NDArray[np.float64],
        "duration": float,
        "temporal_entropy": float
        }
    """
    envelope = amplitude_envelope(data)
    times = np.linspace(0, len(data)/sr, len(data))
    t_q1, t_median, t_q3 = temporal_quartiles(data, sr, kernel_size)

    features = {
        "t_q1": t_q1,
        "t_median": t_median,
        "t_q3": t_q3,
        "temporal_centroid": np.average(times, weights=envelope/np.mean(envelope)),
        "temporal_sd": temporal_sd(data, sr, kernel_size),
        "temporal_skew": skewness(data, sr, kernel_size),
        "temporal_kurtosis": temporal_kurtosis(data, sr, kernel_size),
        "amplitude_envelope": envelope,
        "duration": duration(data, sr, exclude_surrounding_silences=True),
        "temporal_entropy": temporal_entropy(data)[0]
    }

    return features



from pathlib import Path
from typing import Union

try:
    import textgrids
except ModuleNotFoundError:
    raise ImportError("praat-textgrids is required for TextGrid support. Install it with: pip install biosonic[praat]")


def _read_textgrid(
        filepath: Union[str, Path]
    ) -> textgrids.TextGrid:
    """
    """
    filepath = Path(filepath)
    grid = textgrids.TextGrid(filepath)

    return grid



# def modulation_spectrum():
#     """
#     2. Singh NC, Theunissen FE. 2003 Modulation spectra of natural sounds and ethological theories of auditory processing.
#     The Journal of the Acoustical Society of America 114, 3394–3411. (doi:10.1121/1.1624067)
#     """

#     pass



from typing import Any, Literal, Optional, Tuple, Union

import numpy as np
from numpy.typing import ArrayLike
from scipy.signal import butter, filtfilt

from biosonic.compute.utils import hz_to_mel, mel_to_hz


def _check_filterbank_parameters(
        n_filters: int,
        n_fft: int,
        sr: int,
        fmin: float,
        fmax: float,
        ) -> None:

    if fmax > sr / 2:
        raise ValueError(f"fmax must be <= Nyquist frequency (sr/2 = {sr/2}), but got fmax={fmax}")

    if fmin < 0 or fmin >= fmax:
        raise ValueError(f"fmin must be >= 0 and < fmax, but got fmin={fmin}, fmax={fmax}")

    # Ensure n_fft provides enough resolution to resolve unique bins for each filter
    min_bins_required = n_filters + 2
    max_bin_index = int(np.floor((n_fft + 1) * fmax / sr))
    if max_bin_index < min_bins_required:
        raise ValueError(
            f"n_fft={n_fft} doesn't provide enough resolution for {n_filters} filters up to fmax={fmax}Hz "
            f"(only {max_bin_index} frequency bins available). Increase n_fft or reduce fmax/n_filters."
        )


def _filterbank(
        n_filters: int,
        n_fft: int,
        center_freqs: ArrayLike,
        fft_freqs: ArrayLike
    ) -> ArrayLike:
    """
    Construct a triangular filter bank for spectral analysis.

    This function creates a bank of `n_filters` triangular filters to be applied
    to the magnitude spectrum of an audio signal, commonly used in feature
    extraction methods such as Mel-frequency cepstral coefficients (MFCCs).
    Each filter is shaped like a triangle and spans a range of FFT bins
    determined by the `bin_indices` array.

    Parameters
    ----------
    n_filters : int
        Number of filters in the filter bank.

    n_fft : int
        FFT size used to compute the frequency bins. Determines the number of
        frequency bins available.

    bin_indices : ArrayLike
        Array of bin indices (length = n_filters + 2) that specify the start,
        center, and end points of each triangular filter in the filter bank.

    Returns
    -------
    filterbank : ArrayLike
        A 2D array of shape `(n_filters, n_fft // 2 + 1)` containing the filter
        bank matrix. Each row represents a single triangular filter applied
        across the FFT bins.

    Notes
    -----
    The filters are normalized according to the slaney method (directly taken from `librosa`, see references) to
    ensure energy preservation.

    References
    ----------
    McFee, Brian, Colin Raffel, Dawen Liang, Daniel PW Ellis, Matt McVicar, Eric Battenberg,
    and Oriol Nieto. “librosa: Audio and music signal analysis in python.” In Proceedings
    of the 14th python in science conference, pp. 18-25. 2015.
    """
    filterbank = np.zeros((n_filters, int(n_fft // 2 + 1)), dtype=np.float32)

    fdiff = np.diff(center_freqs)
    ramps = np.subtract.outer(center_freqs, fft_freqs)

    for i in range(n_filters):
        # lower and upper slopes
        lower = -ramps[i] / fdiff[i]
        upper = ramps[i + 2] / fdiff[i + 1]

        # intersect with each other and zero
        filterbank[i] = np.maximum(0, np.minimum(lower, upper))

    # normalize each filter to have approx constant energy
    enorm = 2.0 / (center_freqs[2:n_filters+2] - center_freqs[:n_filters])
    filterbank *= enorm[:, np.newaxis]

    # check if any filter is empty
    if np.any(filterbank.max(axis=1) == 0):
        print(
            "Empty filters detected in mel filterbank. "
            "Consider reducing number of bands or increasing n_fft."
        )

    return filterbank


def mel_filterbank(
        n_filters: int,
        n_fft: int,
        sr: int,
        fmin: float = 0.0,
        fmax: Optional[float] = None,
        **kwargs: Any
    ) -> ArrayLike:
    """
    Create a mel spaced triangular filterbank.

    Parameters
    ----------
    n_filters : int
        Number of triangular filters.
    window_length : int
        Window size in samples.
    sr : int
        Sampling rate of the signal in Hz.
    fmin : float
        Minimum frequency in Hz. Must be > 0.
    fmax : float, optional
        Maximum frequency in Hz. Defaults to Nyquist (sr/2).

    Returns
    -------
    filterbank : np.ndarray
        Array of shape (n_filters, n_fft//2 + 1), each row a filter.
    """
    if fmax is None:
        fmax = float(sr) / 2

    _check_filterbank_parameters(n_filters, n_fft, sr, fmin, fmax)

    # fft bin center frequencies
    fft_freqs = np.fft.rfftfreq(n=n_fft, d=1.0 / sr)

    # boundaries
    mel_min = hz_to_mel(fmin, **kwargs)
    mel_max = hz_to_mel(fmax, **kwargs)

    # mel filter center frequencies
    mel_points = np.linspace(mel_min, mel_max, n_filters + 2, dtype=np.float32)
    hz_points = mel_to_hz(mel_points, **kwargs)
    return _filterbank(n_filters, n_fft, hz_points, fft_freqs), mel_points


def linear_filterbank(
        n_filters: int,
        n_fft: int,
        sr: int,
        fmin: float = 0.0,
        fmax: Optional[float] = None
    ) -> ArrayLike:
    """
    Create a linearly spaced triangular filterbank.

    Parameters
    ----------
    n_filters : int
        Number of triangular filters.
    n_fft : int
        FFT size (defines frequency resolution).
    sr : int
        Sampling rate of the signal in Hz.
    fmin : float
        Minimum frequency in Hz. Must be > 0.
    fmax : float, optional
        Maximum frequency in Hz. Defaults to Nyquist (sr/2).

    Returns
    -------
    filterbank : np.ndarray
        Array of shape (n_filters, n_fft//2 + 1), each row a filter.
    """
    if fmax is None:
        fmax = sr / 2

    _check_filterbank_parameters(n_filters, n_fft, sr, fmin, fmax)

    # fft bin center frequencies
    fft_freqs = np.fft.rfftfreq(n=n_fft, d=1.0 / sr)

    hz_points = np.linspace(fmin, fmax, n_filters + 2)

    return _filterbank(n_filters, n_fft, hz_points, fft_freqs), hz_points


def log_filterbank(
    n_filters: int,
    n_fft: int,
    sr: int,
    fmin: float = 0.0,
    fmax: Optional[float] = None,
    base: float = 2.0,
) -> ArrayLike:
    """
    Create a logarithmically spaced triangular filterbank.

    Parameters
    ----------
    n_filters : int
        Number of triangular filters.
    n_fft : int
        FFT size (defines frequency resolution).
    sr : int
        Sampling rate of the signal in Hz.
    fmin : float
        Minimum frequency in Hz. Must be > 0.
    fmax : float, optional
        Maximum frequency in Hz. Defaults to Nyquist (sr/2).
    base : float
        Base of the logarithmic spacing. Typically 2 (for octaves).

    Returns
    -------
    filterbank : np.ndarray
        Array of shape (n_filters, n_fft//2 + 1), each row a filter.
    """
    if fmax is None:
        fmax = sr / 2

    _check_filterbank_parameters(n_filters, n_fft, sr, fmin, fmax)

    if fmin <= 0:
        raise ValueError("fmin must be greater than 0 for log-scaled filterbanks.")

    # fft bin center frequencies
    fft_freqs = np.fft.rfftfreq(n=n_fft, d=1.0 / sr)

    # compute log-spaced center frequencies
    log_min = np.log(fmin) / np.log(base)
    log_max = np.log(fmax) / np.log(base)
    log_points = np.linspace(log_min, log_max, n_filters+2, dtype=np.float32)
    hz_points = base ** log_points

    # convert to FFT bin indices
    bin_indices = np.floor((n_fft + 1) * hz_points / sr).astype(int)
    bin_indices = np.clip(bin_indices, 0, n_fft // 2)

    return _filterbank(n_filters, n_fft, hz_points, fft_freqs), log_points


# TODO weighted filter (seewave), rolloff like in audacity f-filter (tuneR)
def filter(
        data: ArrayLike,
        sr: int,
        f_cutoff: Union[int, Tuple[int, int]],
        type: Literal["lowpass", "highpass", "bandpass", "bandstop"] = "lowpass",
        order: int = 2,
) -> ArrayLike:
    """
    Apply a zero-phase Butterworth filter to a 1D signal using SciPy.

    This function is a wrapper around `scipy.signal.butter` and `scipy.signal.filtfilt`.
    It designs a digital Butterworth filter of the specified type and order,
    then applies it using forward-backward filtering for zero phase distortion.

    Parameters
    ----------
    signal : np.ndarray
        1D input signal to filter.
    sr : int
        Sampling rate of the signal in Hz.
    type : {'lowpass', 'highpass', 'bandpass', 'bandstop'}
        Type of filter to apply. Defaults to 'lowpass'.
    f_cutoff : float or tuple of float
        Cutoff frequency/frequencies in Hz:
        - Single int for 'lowpass' or 'highpass'
        - Tuple of two floats for 'bandpass' or 'bandstop'
        Values must be within (0, Nyquist), where Nyquist = sr / 2.
    order : int
        Filter order. Higher values result in a steeper frequency cutoff,
        but can introduce more edge artifacts and potential instability.
        Defaults to 2, resulting in a slope of 40 dB per decade (i.e. ten-fold change in frequency).

    Returns
    -------
    filtered_signal : np.ndarray
        The filtered signal, same shape as the input.

    Notes
    -----
    - This uses `scipy.signal.butter` and `scipy.signal.filtfilt` to apply the filter forward and backward,
      ensuring zero-phase distortion.

    References
    ----------
    Virtanen P et al. 2020 SciPy 1.0: fundamental algorithms for scientific computing in Python.
    Nat Methods 17, 261–272. (doi:10.1038/s41592-019-0686-2)
    """

    if type in ['bandpass', 'bandstop']:
        if not isinstance(f_cutoff, (list, tuple)) or len(f_cutoff) != 2:
            raise ValueError("f_cutoff must be a tuple/list of two values for bandpass/bandstop filters.")
    else:
        if isinstance(f_cutoff, (list, tuple)):
            raise ValueError("f_cutoff must be a scalar for lowpass/highpass filters.")

    b, a = butter(order, f_cutoff, btype=type, analog=False, fs=sr)
    filtered_signal = filtfilt(b, a, data)
    return filtered_signal



import numpy as np
from numpy.typing import ArrayLike

# def ripple_sound():
#     """
#     Generate a ripple signal ad defined in [2]:

#     Amplitude envelope of each frequency band:
#     .. math::
#         S_i(t,f)=A_icos(2\pi\omega_{t,i}t+2\pi\omega_{f,i}f+\phi_i)

#     Spectrogram:
#     .. math::
#         S(t,f)=A_0+\sum_i S_i(t,f)

#     2. Singh NC, Theunissen FE. 2003 Modulation spectra of natural sounds and ethological theories of auditory processing.
#     The Journal of the Acoustical Society of America 114, 3394–3411. (doi:10.1121/1.1624067)

#     """
#     pass


def amplitude_modulated(
        t: ArrayLike,
        carrier_freq: float,
        modulator_freq: float,
        depth: float = .3
) -> ArrayLike:
    """
    Generate an amplitude-modulated (AM) signal.

    The signal is produced by modulating a carrier sine wave with another sine wave
    (the modulator).

    Parameters
    ----------
    t : ArrayLike
        Time array over which to compute the signal (in seconds).
    carrier_freq : float
        Frequency of the carrier sine wave (Hz).
    modulator_freq : float
        Frequency of the modulating sine wave (Hz).
    depth : float, optional
        Modulation depth (0 to 1), controlling the amplitude variation. Default is 0.3.

    Returns
    -------
    ArrayLike
        Amplitude-modulated signal values at times `t`.
    """
    modulator = (1 - depth) + depth * np.sin(2 * np.pi * modulator_freq * t)
    carrier = np.sin(2 * np.pi * carrier_freq * t)
    return modulator * carrier



from typing import Any, Dict, List, Literal, Optional, Tuple, Union

import numpy as np
from numpy.typing import ArrayLike, NDArray
from scipy import signal
from scipy.fft import fft, ifft, rfft
from scipy.fftpack import dct
from scipy.linalg import solve
from scipy.spatial.distance import cdist

from ..filter import linear_filterbank, log_filterbank, mel_filterbank
from .spectral import power_spectral_entropy
from .temporal import temporal_entropy
from .utils import check_signal_format, check_sr_format, window_signal


def spectrogram(
    data: ArrayLike,
    sr: int,
    window_length: int = 512,
    window: Union[str, ArrayLike] = "hann",
    overlap: float = 50,
    noisereduction: Optional[bool] = False,
    complex_output: bool = False,
) -> Tuple[NDArray[np.float32], NDArray[np.float32], NDArray[np.float32]]:
    """
    Compute the spectrogram of a waveform. Uses scipy.signal.stft and is oriented on seewave.

    Parameters
    ----------
    data : ArrayLike
        Input signal as a 1D array.
    sr : int
        Sampling rate in Hz.
    window_length : int, optional
        Length of the window in samples. Must be even. Default is 512.
    window : str or tuple, optional
        Type of window to use (e.g., 'hann', 'hamming' of scipy.signal.windows) or a custom window array.
        Defaults to 'hann'.
    overlap : float, optional
        Overlap between adjacent windows as a percentage (0–100). Default is 50.
    noisereduction : bool, optional
        Apply noise reduction:
        if True, subtract median from spectrogram values. Default is False.
    complex_output : bool, optional
        If True, return the complex STFT result. If False, return magnitude. Default is False.

    Returns
    -------
    Sx : np.ndarray
        Spectrogram array (complex or magnitude depending on `complex_output`).
    t : np.ndarray
        Time vector corresponding to the columns of `Sx`, in seconds.
    f : np.ndarray
        Frequency vector corresponding to the rows of `Sx`, in Hz.

    Raises
    ------
    ValueError
        If `window_length` is not an even number.

    References
    ----------
    [1] Pauli Virtanen, Ralf Gommers, Travis E. Oliphant, Matt Haberland, Tyler Reddy, David Cournapeau,
    Evgeni Burovski, Pearu Peterson, Warren Weckesser, Jonathan Bright, Stéfan J. van der Walt,
    Matthew Brett, Joshua Wilson, K. Jarrod Millman, Nikolay Mayorov, Andrew R. J. Nelson, Eric Jones,
    Robert Kern, Eric Larson, CJ Carey, İlhan Polat, Yu Feng, Eric W. Moore, Jake VanderPlas, Denis
    Laxalde, Josef Perktold, Robert Cimrman, Ian Henriksen, E.A. Quintero, Charles R Harris, Anne M.
    Archibald, Antônio H. Ribeiro, Fabian Pedregosa, Paul van Mulbregt, and SciPy 1.0 Contributors.
    (2020) SciPy 1.0: Fundamental Algorithms for Scientific Computing in Python. Nature Methods, 17(3),
    261-272. DOI: 10.1038/s41592-019-0686-2.

    [2] J. Sueur, T. Aubin, C. Simonis (2008). “Seewave: a free modular tool for sound analysis and
    synthesis.” Bioacoustics, 18, 213-226.
    """
    if window_length % 2 != 0:
        raise ValueError("'window_length' must be even")

    noverlap = int(window_length * overlap / 100)

    if isinstance(window, str):
        try:
            window = signal.windows.get_window(window, window_length)
        except ValueError as e:
            raise ValueError(f"Invalid window type: {window}") from e
    else:
        window = np.asarray(window)
        if not isinstance(window, np.ndarray):
            raise TypeError("'window' must be either a string or a 1D NumPy array.")

    f, t, Sx = signal.stft(
        data,
        fs=sr,
        window=window,
        nperseg=window_length,
        noverlap=noverlap,
        nfft=window_length,
        padded=False,
        boundary=None
    )

    if complex_output:
        return Sx, t, f

    S_real = np.abs(Sx)

    # Noise reduction
    if noisereduction is True:
        S_real = S_real - np.median(S_real, axis=0)

    return S_real, t, f


def cepstrum(
        data: ArrayLike,
        sr: int,
        mode: Literal["amplitude", "power"] = "amplitude",
    ) -> Tuple[ArrayLike, ArrayLike]:
    """
    Compute the cepstrum of a real-valued time-domain signal.

    The cepstrum is computed by taking the inverse Fourier transform
    of the logarithm of the magnitude spectrum. Depending on the `mode`,
    this function returns either the amplitude or power cepstrum.

    Parameters
    ----------
    data : ArrayLike
        Input real-valued time-domain signal (1D array).
    sr : int
        Sampling rate of the signal in Hz.
    mode : {"amplitude", "power"}, optional
        Type of cepstrum to compute:
        - "amplitude" : Returns the absolute value of the inverse FFT of the log-magnitude spectrum.
        - "power"     : Returns the squared magnitude of the inverse FFT of the log-power spectrum.
        Default is "amplitude".

    Returns
    -------
    cepstrum : ArrayLike
        The computed cepstrum (amplitude or power based on the selected mode).
    quefrencies : ArrayLike
        Array of quefrency values (in seconds), corresponding to each element in the cepstrum.

    References
    ----------
    Childers DG, Skinner DP, Kemerait RC. 1977
    The cepstrum: A guide to processing. Proc. IEEE 65, 1428–1443.
    (doi:10.1109/PROC.1977.10747)
    """
    data = check_signal_format(data)
    sr = check_sr_format(sr)

    if np.all(data == data[0]):
        raise ValueError("Cannot compute cepstrum of flat signal.")

    quefrencies = np.array(range(len(data))) / sr

    if mode == "power":
        return np.abs(ifft(np.log(np.abs(fft(data))**2)))**2, quefrencies

    elif mode == "amplitude":
        return np.abs(ifft(np.log(np.abs(fft(data))))), quefrencies

    else:
        raise ValueError(f"Invalid mode for cepstrum calculation: {mode}")


def cepstral_coefficients(
    data: ArrayLike,
    sr: int,
    window_length: int = 512,
    n_filters: int = 32,
    n_ceps: int = 16,
    pre_emphasis: float = 0.97,
    fmin: float = 0.0,
    fmax: Optional[float] = None,
    filterbank_type: Literal["mel", "linear", "log"] = "mel",
    skip_first: bool = True,
    timestep: float = 0.01,
    **kwargs: Any
) -> ArrayLike:
    """
    Compute cepstral coefficients from a signal using the specified filter bank.

    Parameters
    ----------
    signal : ArrayLike
        Input time-domain signal.
    sr : int
        Sampling rate in Hz.
    window_length : int
        FFT size in samples.
    n_filters : int
        Number of filters in the filter bank.
    n_ceps : int
        Number of cepstral coefficients to return.
    pre_emphasis : float
        Pre-emphasis coefficient to apply. Default is 0.97
    fmin : float
        Minimum frequency for the filter bank.
    fmax : Optional[float]
        Maximum frequency for the filter bank. Defaults to Nyquist (sr/2).
    filterbank_type : {'mel', 'linear', 'log'}
        Type of filter bank to apply before DCT.
    skip_first : bool
        Wether to excluse the first cepstral coefficient. Defaults to True.
    timestep: float
        Step size for window function.
    **kwargs : dict
        Optional keyword arguments for the filter banks, e.g. corner frequency for mel.

    Returns
    -------
    np.ndarray
        Cepstral coefficient array of shape (n_ceps,).
    """
    def liftering(cc: ArrayLike, D: int = 22) -> ArrayLike:
        """
        Apply sinusoidal liftering to cepstral coefficients.

        Parameters:
        - cc: Cepstral coefficients (2D array: frames x coefficients)
        - D: Liftering parameter (default 22)

        Returns:
        - Lifted cepstral coefficients
        """
        cc_lift = np.zeros(cc.shape)

        n = np.arange(1, cc_lift.shape[1] + 1)
        D = 22
        w = 1 + (D / 2) * np.sin(np.pi * n / D)

        return cc * w

    # pre-emphasis - from https://www.geeksforgeeks.org/nlp/mel-frequency-cepstral-coefficients-mfcc-for-speech-recognition/
    data_preemphasized = np.append(data[0], data[1:] - pre_emphasis * data[:-1])

    # window and fft
    data_windowed = window_signal(data_preemphasized, sr, window_length, timestep=timestep)
    mag_frames = np.abs(rfft(data_windowed, window_length))
    pow_frames = (1/window_length) * mag_frames ** 2

    # filter bank selection
    if fmax is None:
        fmax = sr / 2

    if filterbank_type == "mel":
        fbanks, _ = mel_filterbank(n_filters, window_length, sr, fmin, fmax, **kwargs)
    elif filterbank_type == "linear":
        fbanks, _ = linear_filterbank(n_filters, window_length, sr, fmin, fmax)
    elif filterbank_type == "log":
        # raise NotImplementedError("Log frequency scale is not yet implemented for cepstral coefficients in Version 0.")
        fmin_corrected = 1e-6 if fmin == 0 else fmin
        fbanks, _ = log_filterbank(n_filters, window_length, sr, fmin_corrected, fmax, **kwargs)
    else:
        raise ValueError(f"Unknown filterbank_type: {filterbank_type}")

    audio_filtered = np.dot(pow_frames, fbanks.T)
    audio_filtered = np.where(audio_filtered == 0, np.finfo(float).eps, audio_filtered)  # avoid 0 division
    audio_filtered = 20 * np.log10(audio_filtered)  # to dB

    # DCT to cepstral domain
    ceps = dct(audio_filtered, type=2, norm="ortho", axis=1)

    if skip_first:
        ceps = liftering(ceps)[:, 1:n_ceps+1]  # skip C0
    else:
        ceps = liftering(ceps)[:, :n_ceps]

    return np.asarray(ceps.T)


def spectrotemporal_entropy(
        data: ArrayLike,
        sr: int,
        *args: Any,
        **kwargs: Any
    ) -> float:
    """
    Compute the product of temporal_entropy and power spectral entropy of the input data.

    This function computes two information-theoretic measures — temporal temporal_entropy and
    power spectral entropy — using the same unit and multiplies their values
    to produce a combined spectrotemporal complexity measure.

    Parameters
    ----------
    data : ArrayLike
        Input signal as a 1D ArrayLike.
    sr : int
        Sampling rate in Hz.
    unit : {"bits", "nat", "dits", "bans", "hartleys"}, optional
        The logarithmic base to use for temporal_entropy calculations.
        Default is "bits".
    *args : Any
        Additional positional arguments passed to `temporal_entropy` and `power_spectral_entropy`.
    **kwargs : Any
        Additional keyword arguments passed to `temporal_entropy` and `power_spectral_entropy`.

    Returns
    -------
    float
        The temporal_entropy of the input data.

    See Also
    --------
    temporal_entropy : Computes the temporal entropy of the data.
    power_spectral_entropy : Computes the spectral entropy of the data.
    """
    H_t, _ = temporal_entropy(data, *args, **kwargs)
    H_f, _ = power_spectral_entropy(data, sr, *args, **kwargs)
    return H_t * H_f


def dominant_frequencies(
        data: ArrayLike,
        sr: int,
        n_freqs: int = 1,
        min_height: float = 0.05,
        threshold: float = 0.05,
        min_distance: float = 0.05,
        min_prominence: float = 0.05,
        noise_threshold: float = 0.1,
        *args: Any,
        **kwargs: Any
    ) -> NDArray[np.float32]:
    """
    Extracts the dominant frequency or frequencies from each time frame of a spectrogram
    based on the scipy.signal function find_peaks.

    Parameters
    ----------
    data : ArrayLike
        Input 1D audio signal.
    sr : int
        Sample rate of the input signal in Hz.
    n_freqs : Optional[int], default=3
        Number of dominant frequencies to extract per time frame.
        If 1, a 1D array is returned. If >1, a 2D array of shape (time_frames, n_freqs) is returned.
    min_height : Optional[float], default=0.05
        Minimum normalized height of a peak (as a fraction of the spectral magnitude range).
        Must be between 0 and 1.
    min_distance : Optional[float], default=0.05
        Minimum normalized distance between peaks (as a fraction of the total number of frequency bins).
        Must be between 0 and 1.
    min_prominence : Optional[float], default=0.05
        Minimum normalized prominence of a peak (as a fraction of the spectral magnitude range).
        Must be between 0 and 1.
    noise_threshold : Optional[float], default=0.1
        Threshold for a frequency bin to be treated as silent in percent of the median spectrum.
        Must be between 0 and 1.
    *args, **kwargs :
        Additional arguments passed to the scipy.signal ShortTimeFFT class.

    Returns
    -------
    NDArray[np.float32]
        - If n_freqs == 1:
            1D array of shape (time_frames,) containing the dominant frequency per frame.
        - If n_freqs > 1:
            2D array of shape (time_frames, n_freqs) containing the top `n_freqs` dominant
            frequencies per frame. NaNs are used to pad frames with fewer than `n_freqs` detected peaks.
    """
    if not (0.0 <= min_height <= 1.0):
        raise ValueError("min_height must be between 0 and 1")

    if not (0.0 <= min_distance <= 1.0):
        raise ValueError("min_distance must be between 0 and 1")

    if not (0.0 <= min_prominence <= 1.0):
        raise ValueError("min_prominence must be between 0 and 1")

    if not (0.0 <= threshold <= 1.0):
        raise ValueError("threshold must be between 0 and 1")

    if not (0.0 <= noise_threshold <= 1.0):
        raise ValueError("noise_threshold must be between 0 and 1")

    spec, times, freqs = spectrogram(data, sr, *args, **kwargs)
    spec_real = np.abs(spec)

    if n_freqs == 1:
        dominant_freqs = np.full(len(times), np.nan)
    else:
        dominant_freqs = np.full((len(times), n_freqs), np.nan)

    median_range: float = float(np.median([np.max(spec_real[:, t]) - np.min(spec_real[:, t]) for t in range(len(times))]))
    noise_threshold = median_range * noise_threshold

    for t in range(len(times)):
        spectrum = spec_real[:, t]
        magnitude_range = float(np.max(spectrum)) - np.min(spectrum)
        if magnitude_range <= noise_threshold:
            continue

        default_peak_params = {
            "height": magnitude_range*min_height,
            "threshold": magnitude_range*threshold,
            "distance": max(1, len(freqs)//(1/min_distance)),
            "prominence": magnitude_range*min_prominence
        }

        peaks, _ = signal.find_peaks(spectrum, **default_peak_params)

        if len(peaks) > 0:
            sorted_peaks = peaks[np.argsort(spectrum[peaks])[::-1]]
            top_peaks = sorted_peaks[:n_freqs]
            top_freqs = freqs[top_peaks]

            if n_freqs == 1:
                dominant_freqs[t] = top_freqs[0]
            else:
                dominant_freqs[t, :len(top_freqs)] = top_freqs

    return dominant_freqs


def zero_crossings(data: ArrayLike) -> NDArray[np.int64]:
    # TODO
    """
    Calculate the indices of zero crossings in a 1D signal.

    Parameters
    ----------
    data : ArrayLike
        Input 1D audio signal.

    Returns
    -------
    NDArray[np.int64]
        Indices where zero crossings occur.
    """
    pass


def zero_crossing_rate(
        data: ArrayLike,
        frame_length: int = 2048,
        hop_length: int = 512
        ) -> np.float32:
    # TODO
    """
    Calculate the zero crossing rate of a 1D signal.

    Parameters
    ----------
    data : ArrayLike
        Input 1D audio signal.
    frame_length : int, optional
        Length of each frame in samples. Default is 2048.
    hop_length : int, optional
        Number of samples to advance between frames. Default is 512.

    Returns
    -------
    np.float32
        Zero crossing rate of the signal.
    """
    pass


# --------- Tokuda ----------
# https://www.ritsumei.ac.jp/~isao/NLM/

def _variance(x: NDArray[np.float32], dim: int) -> float:
    """Standard deviation of signal after skipping first `dim` points."""
    return float(np.std(x[dim:], ddof=0))


def _nearest_neighbors(
        length: int,
        dim: int,
        ss: NDArray[np.float32],
        c: int,
        exclusion: int
    ) -> Any:
    """Find nearest neighbors for index c based on embedding distance."""
    # Build embedding windows
    target = ss[c - dim + 1:c + 1][None, :]  # shape (1, dim)
    windows = [ss[l - dim + 1:l + 1] for l in range(dim - 1, length - 1) if abs(l - c) > exclusion]
    windows = np.array(windows)

    # Compute distances and sort
    distances = cdist(target, windows, metric="euclidean").ravel()
    sorted_idx = np.argsort(distances)

    # Return indices aligned to original l values
    valid_indices = [l for l in range(dim - 1, length - 1) if abs(l - c) > exclusion]
    return np.array(valid_indices)[sorted_idx]


def lpc_estimate(
        dim: int,
        nnn: int,
        ss: NDArray[np.float32],
        nb: NDArray[np.float32]
    ) -> Any:
    """Estimate LPC coefficients using nearest neighbors (solves Ax = b)."""
    # Cross-correlation vector
    vt = np.array([sum(ss[nb[k] - i] * ss[nb[k] + 1] for k in range(nnn)) for i in range(dim)])

    # Covariance matrix
    A = np.array([[sum(ss[nb[k] - i] * ss[nb[k] - j] for k in range(nnn))
                   for j in range(dim)] for i in range(dim)])

    # Solve A * wd = vt
    try:
        wd = solve(A, vt, assume_a="sym")  # covariance is symmetric
    except np.linalg.LinAlgError:
        wd = np.zeros(dim)  # fallback: predict nothing
    return wd


def SNR(
        length: int,
        dim: int,
        nnn: int,
        xx: NDArray[np.float32],
        exclusion: int
    ) -> float:
    ss = xx.copy()
    rr = np.zeros_like(ss)

    # Predict each point using local linear model
    for i in range(dim - 1, length - 1):
        rr[i + 1] = ss[i + 1]
        nb = _nearest_neighbors(length, dim, ss, i, exclusion)
        wd = lpc_estimate(dim, nnn, ss, nb)
        rr[i + 1] -= np.dot(wd, ss[i - dim + 1:i + 1][::-1])

    # Variance of original and residual signals
    vs = _variance(ss, dim)
    vr = _variance(rr, dim)
    return float(10.0 * np.log10(vs / vr))


def tokuda_nlm(
        dim: int,
        data: NDArray[np.float32],
        exclusion: int = 15
        ) -> Tuple[List[Tuple[float, float]], float, float]:
    """
    Compute the DVS plot values and the nonlinear measure (NLM).

    Parameters
    ----------
    dim : int
        Embedding dimension.
    data : array-like
        Time series data (1D).
    exclusion : int, default=15
        Minimum temporal separation for nearest neighbors.

    Returns
    -------
    results : list of tuple of (float, float)
        List of (percentage of neighbors, SNR) pairs.
    nlm_value : float
        Nonlinear measure in dB (difference between max SNR and full-data SNR).
    """
    xx = np.asarray(data, dtype=float)
    length = len(xx)
    if length <= 2 * exclusion + dim + 1:
        raise ValueError("Time series too short for given dim and exclusion.")

    # step size for sweeping nearest neighbor counts
    step = max(1, int(0.025 * length))

    results: List[Tuple[float, float]] = []
    max_snr: Optional[float] = None

    # Sweep number of neighbors from dim+1 up to max allowed
    for nnn in range(dim + 1, length - dim - 2 * exclusion, step):
        snr = SNR(length, dim, nnn, xx, exclusion)
        p_nnn = 100.0 * nnn / length  # percentage of neighbors
        results.append((p_nnn, snr))
        if max_snr is None or snr > max_snr:
            max_snr = snr

    # Fallback if loop produced no results
    if max_snr is None:
        return [], float("nan"), float("nan")

    # Compute SNR using all admissible neighbors (upper bound)
    nnn = length - dim - 2 * exclusion - 1
    snr_all = SNR(length, dim, nnn, xx, exclusion)

    # Nonlinear measure = difference between maximal SNR and full-data SNR
    nlm_value = max_snr - snr_all

    return results, nlm_value, snr_all

# ----------- Tokuda End --------------

# TODO Modulation spectra


def calculate_dominant_frequency_features(
        data: ArrayLike,
        sr: int,
        **kwargs: Any
    ) -> Dict[str, Union[float, NDArray[np.float32]]]:
    """
    Calculate dominant frequency features.
    """
    dominant_freqs = dominant_frequencies(data, sr, n_freqs=1, **kwargs)

    # exclude 0 values (no peak detected) from calculations
    dom_freqs_detected = dominant_freqs[dominant_freqs > 0]
    min_dom: float
    max_dom: float
    min_dom, max_dom = float(np.min(dom_freqs_detected)), float(np.max(dom_freqs_detected))
    range_dom = max_dom - min_dom
    cumulative_diff: float = np.sum(np.abs(np.diff(dom_freqs_detected)))
    mod_dom = cumulative_diff / range_dom if range_dom > 0 else 0

    return {
        "mean_dom": np.mean(dom_freqs_detected),
        "min_dom": min_dom,
        "max_dom": max_dom,
        "range_dom": range_dom,
        "mod_dom": mod_dom
    }


def spectrotemporal_features(
        data: ArrayLike,
        sr: int,
        n_dominant_freqs: int = 1,
        **kwargs: Any
    ) -> dict[str, Union[float, np.floating, NDArray[np.float32]]]:
    """
    Extracts a set of spectrotemporal features from a signal.

    Args:
        data : ArrayLike
            The input signal as a 1D ArrayLike.
        sr : int
            Sampling rate of the signal in Hz.
        n_dominant_frequencies : int
            Number of dominant frequencies to extract. Default is 1.
        **kwargs : dict[str, Any]
            Optional parameters for dominant frequency estimation.

    Retuns:
        dict
        {"spectrotemporal_entropy": float,
        "dominant_frequencies": ArrayLike}
    """
    data = check_signal_format(data)
    check_sr_format(sr)
    features = {
        "spectrotemporal_entropy": spectrotemporal_entropy(data, sr),
        "dominant_freqs": dominant_frequencies(data, sr, n_freqs=n_dominant_freqs, **kwargs),
    }

    dom_freq_feats = calculate_dominant_frequency_features(data, sr, **kwargs)

    return {**features, **dom_freq_feats}



# from dataclasses import dataclass
import traceback
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Tuple, Union, get_args

import numpy as np
import pandas as pd
from numpy.typing import ArrayLike, NDArray
from scipy.io import wavfile
from scipy.signal import resample

QuantizationStr = Literal["int8", "int16", "int32", "float32", "float64"]

# @dataclass
# class Signal:
#     """
#     A dataclass representing an audio signal.

#     Attributes:
#         data : NDArray
#             Audio samples as a NumPy array. 1D for mono, otherwise a 2D array with shape (n_samples, n_channels)
#         n_channels : int
#             Number of audio channels (1 for mono, 2 for stereo, etc.).
#         sr : int
#             Sample rate in Hz.
#         quantization : QuantizationStr
#             Data format (bit depth) of the signal.
#     """
#     data: NDArray
#     n_channels: int
#     sr: int
#     quantization: QuantizationStr


def convert_dtype(data: NDArray, target_dtype: QuantizationStr) -> NDArray:
    """
    Converts audio data to the specified quantization format.

    Parameters
    ----------
        data : NDArray
            Input audio data.
        target_dtype : QuantizationStr
            Target NumPy dtype as a string. Options are: "int8", "int16", "int32", "float32", "float64"

    Returns
    -------
        NDArray
            Converted audio data with appropriate scaling and clipping.

    Notes
    -----

    The table below summarizes the value ranges and corresponding NumPy dtypes
    for common WAV audio formats. See the SciPy io.wavfile documentation for more details.

    +------------------------+--------------+---------------+-------------+
    | WAV format             | Min          | Max           | NumPy dtype |
    +========================+==============+===============+=============+
    | 32-bit floating-point  | -1.0         | +1.0          | float32     |
    +------------------------+--------------+---------------+-------------+
    | 32-bit integer PCM     | -2147483648  | +2147483647   | int32       |
    +------------------------+--------------+---------------+-------------+
    | 24-bit integer PCM     | -2147483648  | +2147483392   | int32       |
    +------------------------+--------------+---------------+-------------+
    | 16-bit integer PCM     | -32768       | +32767        | int16       |
    +------------------------+--------------+---------------+-------------+
    | 8-bit integer PCM      | 0            | 255           | uint8       |
    +------------------------+--------------+---------------+-------------+

    References
    ----------
    Virtanen P et al. 2020 SciPy 1.0: fundamental algorithms for scientific computing in Python.
    Nat Methods 17, 261–272. (doi:10.1038/s41592-019-0686-2)
    """
    if target_dtype not in get_args(QuantizationStr):
        raise ValueError(f"Invalid quantization: {target_dtype}. Must be one of {get_args(QuantizationStr)}")

    target_np_dtype: np.dtype[np.generic] = np.dtype(target_dtype)
    current_dtype = data.dtype

    if current_dtype == target_np_dtype:
        return data

    # special handling for uint8 (unsigned), to float32
    if current_dtype == np.uint8 and np.issubdtype(target_np_dtype, np.floating):
        return ((data.astype(target_np_dtype) - 128) / 128).astype(target_np_dtype)

    if np.issubdtype(current_dtype, np.floating) and target_np_dtype == np.uint8:
        return ((data * 128) + 128).clip(0, 255).astype(np.uint8)

    # integer to float: normalize
    if np.issubdtype(current_dtype, np.integer) and np.issubdtype(target_np_dtype, np.floating):
        max_val = np.iinfo(current_dtype).max
        return (data.astype(np.float32) / max_val).astype(target_np_dtype)

    # float to integer: scale and clip
    if np.issubdtype(current_dtype, np.floating) and np.issubdtype(target_np_dtype, np.integer):
        max_val = np.iinfo(target_np_dtype).max
        return (data * max_val).clip(-max_val, max_val - 1).astype(target_np_dtype)

    # integer to integer or float to float
    return data.astype(target_np_dtype)


def resample_audio(data: NDArray, orig_sr: int, target_sr: int) -> NDArray:
    """
    Resample audio data from an original sampling rate to a target sampling rate.

    Parameters
    ----------
    data : NDArray
        Input audio data array. Shape can be (n_samples,) for mono or (n_samples, n_channels) for multi-channel audio.
    orig_sr : int
        Original sample rate of the audio data in Hz.
    target_sr : int
        Desired sample rate in Hz.

    Returns
    -------
    NDArray
        Resampled audio data with shape adjusted to the target sample rate.
        Number of channels remains unchanged.

    Notes
    -----
    - If `orig_sr` is equal to `target_sr`, the original data is returned unchanged.
    - Uses `scipy.signal.resample` internally for resampling each channel independently.

    References
    ----------
    Virtanen P et al. 2020 SciPy 1.0: fundamental algorithms for scientific computing in Python.
    Nat Methods 17, 261–272. (doi:10.1038/s41592-019-0686-2)
    """
    if orig_sr == target_sr:
        return data
    n_samples = round(data.shape[0] * target_sr / orig_sr)
    if data.ndim == 1:
        return resample(data, n_samples)
    else:
        return np.stack([resample(data[:, ch], n_samples) for ch in range(data.shape[1])], axis=1)


def convert_channels(data: NDArray, target_channels: int) -> NDArray:
    """
    Convert audio data to a target number of channels (mono or stereo).

    Parameters
    ----------
    data : NDArray
        Input audio data array. Shape is (n_samples,) for mono or (n_samples, n_channels) for multi-channel.
    target_channels : int
        Desired number of channels. Supported values are 1 (mono) or 2 (stereo).

    Returns
    -------
    NDArray
        Audio data converted to the target number of channels.

    Raises
    ------
    NotImplementedError
        If `target_channels` is greater than 2.

    Notes
    -----
    - Converts stereo to mono by averaging channels.
    - Converts mono to stereo by duplicating the mono channel.
    - If the data already has the target number of channels, it is returned unchanged.
    """
    if target_channels > 2:
        raise NotImplementedError("Conversion to more than 2 channels not implemented yet.")

    n_ch = 1 if data.ndim == 1 else data.shape[1]

    if n_ch == target_channels:
        return data

    if target_channels == 1:
        # Convert stereo to mono by averaging channels
        return data.mean(axis=1)

    elif target_channels == 2:
        # Convert mono to stereo by duplicating
        return np.stack([data, data], axis=1)

    return data


def read_wav(
        filepath: Union[str, Path],
        sampling_rate: Optional[int] = None,
        quantization: QuantizationStr = "float32",
        n_channels: Optional[int] = None,
    ) -> Tuple[NDArray, int, int, QuantizationStr]:
    """
    Reads a WAV file and returns a Signal object, optionally converting sample rate, number of channels,
    and quantization format.

    Parameters
    ----------
    filepath : str or Path
        Path to the WAV file to read.
    sampling_rate : int, optional
        Target sample rate in Hz. If specified and different from the file's original sample rate,
        resampling should be applied (using scipy.signal.resample - Fourier method).
        If None, the original sample rate is kept.
    quantization : {'int8', 'int16', 'int32', 'float32', 'float64'}, optional
        Desired output data format for the audio samples (default is "float32").
    n_channels : int, optional
        Target number of audio channels (e.g., 1 for mono, 2 for stereo).
        If specified and different from the original number of channels, conversion
        is be applied (downmixing from stereo to mono (np.mean) and copying from mono to stereo).
        If None, the original number of channels is kept.

    Returns
    -------
    A tuple containing:
        - data : np.ndarray
            Audio samples, shape (n_samples,) for mono or (n_samples, n_channels) otherwise.
        - sr : int
            Sample rate in Hz.
        - n_channels : int
            Number of audio channels.
        - quantization : str
            Data format of the signal.

    Notes
    -----
    - Uses `scipy.io.wavfile` to read WAV files.
    - 24-bit WAV files are stored as `np.int32` (as in `scipy.io.wavfile`).

    References
    ----------
    Virtanen P et al. 2020 SciPy 1.0: fundamental algorithms for scientific computing in Python.
    Nat Methods 17, 261–272. (doi:10.1038/s41592-019-0686-2)
    """
    sr, data = wavfile.read(filepath)

    if sampling_rate is not None:
        if sr != sampling_rate:
            data = resample_audio(data, sr, sampling_rate)
            sr = sampling_rate

    n_ch = 1 if data.ndim == 1 else data.shape[1]
    if n_channels is not None:
        if n_channels != n_ch:
            data = convert_channels(data, n_channels)
            n_ch = n_channels

    if quantization != data.dtype.name:
        data = convert_dtype(data, quantization)

    return data, sr, n_ch, quantization


def batch_normalize_wav_files(
    folder_path: Union[str, Path],
    target_sr: int,
    target_channels: int,
    target_quantization: QuantizationStr,
    output_dir: Optional[Union[str, Path]] = None
) -> None:
    """
    Batch normalize all WAV files in a folder to the same sample rate, number of channels, and quantization.

    Parameters
    ----------
    folder_path : str or Path
        Path to the folder containing input WAV files to normalize.
    target_sr : int
        Target sample rate in Hz to convert all audio files to.
    target_channels : int
        Target number of audio channels (e.g., 1 for mono, 2 for stereo).
    target_quantization : {'int8', 'int16', 'int32', 'float32', 'float64'}
        Target bit depth / data format for the output audio files.
    output_dir : str or Path, optional
        Directory to save the normalized WAV files. If None, a 'normalized' subfolder
        will be created inside `folder_path`.

    Returns
    -------
    None
        This function saves the normalized WAV files to disk and does not return anything.

    Notes
    -----
    - Input WAV files with extensions '.wav' and '.WAV' are processed.
    - Uses `read_wav` for loading and converting audio files. This attaches to scipys wavfile.io.read function.
    - Output files are saved with the same filename in the output directory. So if you set output_dir to your folder_path,
      **all origninal files will be overwritten!**
    """
    folder_path = Path(folder_path)
    output_dir = Path(output_dir) if output_dir else folder_path / "normalized"
    output_dir.mkdir(parents=True, exist_ok=True)

    wav_files = list(folder_path.glob("*.wav")) + list(folder_path.glob("*.WAV"))
    for wav_file in wav_files:
        data, _, _, _ = read_wav(wav_file, target_sr, target_quantization, target_channels)
        out_path = output_dir / wav_file.name
        wavfile.write(out_path, target_sr, data.astype(np.dtype(target_quantization)))
        print(f"Normalized: {wav_file.name} -> {out_path}")


def batch_extract_features(
        folder_path: Union[str, Path],
        save_csv_path: Optional[str] = None
) -> pd.DataFrame:
    """
    Extract features from all WAV files in a folder.

    This function processes all `.wav` and `.WAV` files in the specified folder,
    applies feature extraction to each file using `extract_all_features`, and
    returns a DataFrame containing the extracted features. Optionally, the
    results can be saved to a CSV file.

    Parameters
    ----------
    folder_path : str or pathlib.Path
        Path to the folder containing `.wav` audio files.

    save_csv_path : str, optional
        Path to save the resulting feature DataFrame as a CSV file.
        If None (default), the CSV is not saved.

    Returns
    -------
    pandas.DataFrame
        A DataFrame where each row contains features extracted from one WAV file.
        Includes a `filename` column with the original file name.

    Raises
    ------
    This function handles file-level errors internally and prints tracebacks,
    but it does not raise exceptions during feature extraction or saving.

    Notes
    -----
    - This function relies on `read_wav` and `extract_all_features`.
    - Files that fail to process are skipped, and their errors are printed.
    - All features are returned in a single DataFrame.
    """
    from biosonic.compute.utils import extract_all_features

    folder_path = Path(folder_path)
    feature_rows = []

    wav_files = list(folder_path.glob("*.wav")) + list(folder_path.glob("*.WAV"))
    for wav_file in wav_files:
        print(f"processing {wav_file.name}")
        try:
            data, sr, _, _ = read_wav(wav_file)
            features = extract_all_features(data, sr)
            features['filename'] = wav_file.name
            feature_rows.append(features)
        except Exception as e:
            print(f"Failed to process {wav_file.name}: {e}")
            traceback.print_exc()

    out_df = pd.DataFrame(feature_rows)

    if save_csv_path:
        try:
            out_df.to_csv(folder_path/save_csv_path, index=False)
            print(f"Features saved to: {folder_path/save_csv_path}")
        except Exception as e:
            print(f"Failed to save CSV: {e}")

    return out_df


def batch_read_files_to_df(
        folder_path: Union[str, Path],
        save_csv_path: Optional[str] = None,
) -> pd.DataFrame:
    """
    Extract features from all WAV files in a folder.

    This function processes all `.wav` and `.WAV` files in the specified folder,
    and returns a DataFrame containing the waveforms and filenames. Optionally, the
    results can be saved to a CSV file.

    Parameters
    ----------
    folder_path : str or pathlib.Path
        Path to the folder containing `.wav` audio files.

    save_csv_path : str, optional
        Path to save the resulting DataFrame as a CSV file.
        If None (default), the CSV is not saved.

    Returns
    -------
    pandas.DataFrame
        A DataFrame where each row contains waveform, sample rate and filename from one WAV file.

    Raises
    ------
    This function handles file-level errors internally and prints tracebacks,
    but it does not raise exceptions during feature extraction or saving.

    Notes
    -----
    - This function relies on `read_wav`.
    - Files that fail to process are skipped, and their errors are printed.
    """
    folder_path = Path(folder_path)
    rows = []

    wav_files = list(folder_path.glob("*.wav")) + list(folder_path.glob("*.WAV"))
    for wav_file in wav_files:
        print(f"processing {wav_file.name}")
        try:
            data, sr, _, _ = read_wav(wav_file)
            columns: dict[str, Any] = {}
            columns['filename'] = wav_file.name
            columns['sr'] = sr
            columns['waveform'] = data
            rows.append(columns)
        except Exception as e:
            print(f"Failed to process {wav_file.name}: {e}")
            traceback.print_exc()

    out_df = pd.DataFrame(rows)

    if save_csv_path:
        try:
            out_df.to_csv(save_csv_path, index=False)
            print(f"Features saved to: {save_csv_path}")
        except Exception as e:
            print(f"Failed to save CSV: {e}")

    return out_df


def segments_from_signal(
        data: NDArray,
        sr: int,
        boundaries: Union[Dict[str, float], ArrayLike, Tuple[float, float], List[Dict[str, float]]]
    ) -> List[NDArray]:
    """
    Extract segments from an audio signal based on time boundaries.

    Parameters
    ----------
    data : NDArray
        The audio signal array.
    sr : int
        Sampling rate of the audio signal.
    boundaries : Union[Dict[str, float], Tuple[float, float], ArrayLike, List[Dict[str, float]]]
        Segment boundaries. Supported formats:
        - Single dict with 'begin' and 'end' keys
        - Tuple of (begin, end)
        - 2D ArrayLike of shape (n, 2), each row as (begin, end)
        - List of dicts with 'begin' and 'end' keys

    Returns
    -------
    List[NDArray]
        A list of segmented portions of the audio signal.
    """
    # TODO handle multiple channels
    segments = []

    # Normalize boundaries to a list of (begin, end) tuples
    if isinstance(boundaries, dict):
        boundaries = [(boundaries["begin"], boundaries["end"])]
    elif isinstance(boundaries, tuple) and len(boundaries) == 2:
        boundaries = [boundaries]
    elif isinstance(boundaries, list) and all(isinstance(b, dict) for b in boundaries):
        boundaries = [(b["begin"], b["end"]) for b in boundaries]
    elif isinstance(boundaries, (np.ndarray, list)):
        boundaries = list(boundaries)  # ensure list of lists or tuples
    else:
        raise ValueError("Unsupported boundary format")

    for begin, end in boundaries:
        start_idx = int(np.floor(begin * sr))
        end_idx = int(np.ceil(end * sr))
        segments.append(data[start_idx:end_idx])

    return segments


def boundaries_from_textgrid(
        filepath: Union[str, Path],
        tier_name: str
    ) -> List[Dict[str, float]]:
    """
    Extracts segment boundaries from a specified tier in a praat TextGrid file.

    Parameters
    ----------
    filepath : Union[str, Path]
        Path to the TextGrid file.
    tier_name : str
        The name of the tier from which to extract intervals.

    Returns
    -------
    List[Dict[str, float]]
        A list of dictionaries, each containing 'begin', 'end', and 'label' keys for non-empty intervals in the specified tier.
    """

    try:
        from biosonic.praat import _read_textgrid
    except ImportError:
        raise ImportError("praat-textgrids is required for TextGrid support. Install it with: pip install biosonic[praat]")

    grid = _read_textgrid(filepath)
    segments = grid.interval_tier_to_array(tier_name=tier_name)
    return [segment for segment in segments if len(segment["label"]) > 0]


def boundaries_from_raven(
        filepath: Union[str, Path]
    ) -> List[Dict[str, float]]:
    """
    Extracts one segment per selection from a Raven selection table (.txt).

    Parameters
    ----------
    filepath : str or Path
        Path to the Raven selection table file.

    Returns
    -------
    List[Dict[str, float]]
        A list of dicts, each containing:
        - 'begin': start time in seconds
        - 'end': end time in seconds
        - 'label': merged annotation text from all rows of that selection
    """
    df = pd.read_csv(filepath, sep="\t")

    # Group by selection number, merge times & annotations
    grouped = (
        df.groupby("Selection")
          .agg({
              "Begin Time (s)": "min",
              "End Time (s)": "max",
              "Annotation": lambda ann: "; ".join(
                  sorted({str(v).strip() for v in ann if pd.notna(v) and str(v).strip()})
              )
            })
          .reset_index()
    )

    boundaries = []
    for _, row in grouped.iterrows():
        boundaries.append({
            "begin": float(row["Begin Time (s)"]),
            "end": float(row["End Time (s)"]),
            "label": row["Annotation"]
        })
    return boundaries


def audio_segments_from_textgrid(
        data: NDArray,
        sr: int,
        filepath_textgrid: Union[str, Path],
        tier_name: str,
        as_df: bool = True,
        **kwargs: Any
    ) -> Union[pd.DataFrame, list[dict[str, Any]]]:
    """
    Extracts and visualizes audio segments corresponding to labeled intervals
    in a praat TextGrid file.

    Parameters
    ----------
    data : NDArray
        The audio signal array.
    sr : int
        Sampling rate of the audio signal.
    filepath_textgrid : Union[str, Path]
        Path to the TextGrid file containing segmentation information.
    tier_name : str
        Name of the tier to extract labeled segments from.
    as_df : Optional[bool]
        Whether to return as pandas DataFrame or list of dictionaries. Defaults to true.

    Returns
    -------
    List[Dict[NDArray, str]] or DataFrame
        A list of dictionaries, each containing:
        - The audio segment (NDArray) for each labeled interval.
        - The corresponding label (str).
        Or a homologue dataframe, additionally holding sr and filepath.
    """
    from biosonic.plot import plot_boundaries_on_spectrogram

    filepath_textgrid = Path(filepath_textgrid)

    boundaries = boundaries_from_textgrid(filepath_textgrid, tier_name)

    plot_boundaries_on_spectrogram(data, sr, boundaries, **kwargs)
    segments = segments_from_signal(data, sr, boundaries)
    if as_df:
        return pd.DataFrame([{
                "waveform": seg,
                "label": str(b["label"]),
                "sr": sr,
                "filename": str(filepath_textgrid.name)
            } for seg, b in zip(segments, boundaries)])
    return [{"waveform": seg, "label": str(b["label"])} for seg, b in zip(segments, boundaries)]


def audio_segments_from_raven(
        data: NDArray,
        sr: int,
        filepath_raven: Union[str, Path]
    ) -> List[Dict[NDArray, str]]:
    """
    Extracts and visualizes audio segments corresponding to intervals
    in a Raven selection table.

    Parameters
    ----------
    data : NDArray
        The audio signal array.
    sr : int
        Sampling rate of the audio signal.
    filepath_raven : Union[str, Path]
        Path to the txt file containing segmentation information.

    Returns
    -------
    List[Dict[NDArray, str]]
        A list of dictionaries, each containing:
        - The audio segment (NDArray) for each labeled interval.
        - The corresponding label (str).
    """
    from biosonic.plot import plot_boundaries_on_spectrogram

    boundaries = boundaries_from_raven(filepath_raven)

    plot_boundaries_on_spectrogram(data, sr, boundaries)
    segments = segments_from_signal(data, sr, boundaries)
    return [{"data": seg, "label": str(b["label"])} for seg, b in zip(segments, boundaries)]



import warnings
from typing import Any, Dict, Literal, Optional, Tuple, Union

import numpy as np
from numpy.typing import ArrayLike, NDArray
from scipy import fft
from scipy.stats import gmean

from .utils import (
    check_signal_format,
    check_sr_format,
    cumulative_distribution_function,
    exclude_trailing_and_leading_zeros,
    shannon_entropy
)


def spectrum(data: ArrayLike,
             sr: Optional[int] = None,
             mode: Union[str, int, float] = 'amplitude') -> Tuple[Optional[NDArray[np.float32]], NDArray[np.float32]]:
    """
    Computes the magnitude spectrum of a signal, allowing for amplitude, power, or arbitrary exponentiation of the magnitude.

    Parameters
    ----------
        data : ArrayLike
            The input time-domain signal as a 1D array-like.
        sr: Optional Integer, default=None
            Sampling rate in Hz as an integer. If given, returns the frequency bins
            of the magnitude spectrum. Defaults to None.
        mode : Union[str, int], default='amplitude'
            Specifies how to compute the spectrum:
            - 'amplitude': return the amplitude spectrum (|FFT|).
            - 'power': return the power spectrum (|FFT|^2).
            - int or float: raise the magnitude to the given power (e.g., 3 for |FFT|^3).

    Returns
    -------
        Tuple[NDArray[np.float64], NDArray[np.float64]]
            A tuple (frequencies, spectrum), where:
            - frequencies: If sr is provided. 1D array of frequency bins corresponding to the spectrum.
            - spectrum: 1D array of the transformed frequency-domain representation (magnitude raised to the specified power).

    Raises
    ------
        ValueError
            If `mode` is a string but not one of the supported options.
        TypeError
            If `mode` is not a string, int, or float.
    """
    data = check_signal_format(data)

    freqs = None

    if data.size == 0:
        warnings.warn("Input signal is empty; returning an empty spectrum.", RuntimeWarning)
        return freqs, np.array([], dtype=np.float32)

    if sr is not None:
        sr = check_sr_format(sr)
        freqs = fft.rfftfreq(len(data), d=1/sr)

    magnitude_spectrum = np.abs(fft.rfft(data))

    if isinstance(mode, str):
        mode = mode.lower()
        if mode == 'amplitude':
            return freqs, magnitude_spectrum
        elif mode == 'power':
            return freqs, magnitude_spectrum ** 2
        else:
            raise ValueError(f"Invalid string mode '{mode}'. Use 'amplitude', 'power', or an integer.")
    elif isinstance(mode, (int, float)) and not isinstance(mode, bool):
        return freqs, magnitude_spectrum ** mode
    else:
        raise TypeError(f"'mode' must be a string, int or float, not {type(mode).__name__}.")


def quartiles(data: ArrayLike, sr: int) -> Tuple[float, float, float]:
    """
    Compute the 1st, 2nd (median), and 3rd quartiles of the power spectrum of a signal.

    Parameters
    ----------
        data : ArrayLike
            Input signal as a 1D array-like.
        sr : int
            Sampling rate (in Hz).

    Returns
    -------
        q1 : float
            Frequency at which the cumulative power spectrum reaches the 25% mark.
        q2 : float
            Frequency at which the cumulative power spectrum reaches the 50% mark (median frequency).
        q3 : float
            Frequency at which the cumulative power spectrum reaches the 75% mark.

    Raises
    ------
        ValueError
            If the input signal is empty.
        ValueError
            If the input signal contains only zeros.
        ValueError
            If the spectrum output is inconsistent (e.g., mismatched lengths).

    Notes
    -----
        This function calculates spectral quartiles using the cumulative distribution function (CDF)
        of the signal's power spectrum. Quartiles are determined by finding the frequencies
        at which the CDF crosses 25%, 50%, and 75%.

    See Also
    --------
        spectrum : Computes the spectral envelope of a signal.
        cumulative_distribution_function : Computes the normalized cumulative sum of a spectrum.

    Examples
    --------
        >>> import numpy as np
        >>> from mymodule import spectral_quartiles
        >>> sr = 1000
        >>> t = np.linspace(0, 1, sr, endpoint=False)
        >>> x = np.sin(2 * np.pi * 10 * t) + 0.5 * np.sin(2 * np.pi * 30 * t)
        >>> spectral_quartiles(x, sr)
        (10.0, 20.0, 30.0)
    """
    data = check_signal_format(data)
    sr = check_sr_format(sr)
    if len(data) == 0:
        raise ValueError("Input is empty")
    if np.all(data == 0):
        raise ValueError("Signal contains no nonzero values")

    frequencies, envelope = spectrum(data, sr=sr, mode="power")
    cdf = cumulative_distribution_function(envelope)

    if frequencies is None or len(frequencies) != len(envelope):
        raise ValueError("Freuency bins don't match envelope")

    return frequencies[np.searchsorted(cdf, 0.25)], frequencies[np.searchsorted(cdf, 0.5)], frequencies[np.searchsorted(cdf, 0.75)]


def flatness(data: ArrayLike) -> Union[float, np.floating[Any]]:
    """
    Compute the spectral flatness (also known as Wiener entropy) of a signal.

    Spectral flatness is a measure of how noise-like a signal is. A flatness close to 1
    indicates a flat (white noise-like) spectrum, whereas a value close to 0 indicates
    a peaky (tonal) spectrum.

    Parameters
    ----------
        data : ArrayLike
            Time-domain input signal (1D array-like).

    Returns
    -------
        float
            Spectral flatness value, defined as the ratio of the geometric mean to the arithmetic mean
            of the signal's power spectrum (excluding leading/trailing zeros).

    Raises
    ------
        ValueError
            If input signal is empty, contains only zeros, or the computed flatness is not a float (from np.mean and scipy's gmean).

    References
    ----------
        Sueur, J. (2018). Sound Analysis and Synthesis with R*. Springer International Publishing, p. 299.
        https://doi.org/10.1007/978-3-319-77647-7
    """
    data = check_signal_format(data)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        _, envelope = spectrum(data, mode="power")

    ps_wo_zeros = exclude_trailing_and_leading_zeros(envelope)
    if len(ps_wo_zeros) == 0:
        raise ValueError("Input signal contained only zero values")

    flatness_ = gmean(ps_wo_zeros) / np.mean(ps_wo_zeros)
    if not np.isscalar(flatness_) or not isinstance(flatness_, (float, np.floating)):
        raise ValueError(f"Received wrong data type for spectral flatness: {type(flatness_)}")
    # TODO transform to dB
    return flatness_


def spectral_moments(
        data: ArrayLike,
        sr: int
        ) -> Tuple[
            Union[float, np.floating[Any]],
            Union[float, np.floating[Any]],
            Union[float, np.floating[Any]],
            Union[float, np.floating[Any]]
            ]:
    """
    Calculate the first four spectral moments.

    Parameters
    ----------
    data : ArrayLike
        1D array-like representing the input signal.
    sampling_rate : float
        Sampling rate of the signal in Hz.

    Returns
    -------
        float or np.floating
            Spectral centroid in Hz
        float or np.floating
            Spectral bandwidth in Hz
        float or np.floating
            Spectral skewness
        float or np.floating
            Spectral kurtosis


    References
    ----------
        Klapuri A, Davy M. 2006 Signal processing methods for music transcription.
        New York: Springer. p.136
    """
    data = check_signal_format(data)
    sr = check_sr_format(sr)

    if np.all(data == 0):
        raise ValueError("Signal contains no nonzero values")

    freqs, ms = spectrum(data, sr=sr)
    # normalize spectrum
    ms = ms / np.sum(ms)
    centroid_ = np.average(freqs, weights=ms)
    bandwidth_ = np.sqrt(np.sum(ms * (freqs-centroid_)**2))
    if bandwidth_ == 0:
        warnings.warn("Bandwidth of signal is 0, returning NaN for skewness and kurtosis", RuntimeWarning)
        return centroid_, bandwidth_, np.nan, np.nan
    skewness_ = (np.sum(ms * (freqs-centroid_)**3))/(bandwidth_**3)
    kurtosis_ = (np.sum(ms * (freqs-centroid_)**4))/(bandwidth_**4)

    return centroid_, bandwidth_, skewness_, kurtosis_


def centroid(data: ArrayLike, sr: int) -> Union[float, np.floating[Any]]:
    r"""
    Compute the spectral centroid of a signal.

    The spectral centroid represents the "center of mass" of the power spectrum,
    giving a measure of where the energy of the spectrum is concentrated. It is
    calculated as the first spectral moment, or the weighted average of the frequency components, using the
    magnitude spectrum as weights:

        .. math::
            C_f=\sum_{k \epsilon K_+}k X(k)

    Parameters
    ----------
        data : ArrayLike
            Input time-domain signal. Must be one-dimensional and convertible to a NumPy array.

        sr : int
            Sampling rate of the input signal in Hz.

    Returns
    -------
        float or np.floating
            Spectral centroid in Hz. A higher value indicates that the signal's energy
            is biased toward higher frequencies.

    Raises
    ------
        ValueError
            If the computed centroid is not a scalar floating-point value.

    Examples
    --------
        >>> signal = np.random.randn(1024)
        >>> sr = 44100
        >>> centroid(signal, sr)
        7101.56

    References
    ----------
        Klapuri A, Davy M. 2006 Signal processing methods for music transcription.
        New York: Springer. p.136
    """
    centroid_, _, _, _ = spectral_moments(data, sr)

    return centroid_


def bandwidth(data: ArrayLike, sr: int) -> Union[float, np.floating[Any]]:
    r"""
    Compute the mean spectral bandwidth (standard deviation or second spectral moment) of a signal.
    It is calculated as

        .. math::
            S_f=S_f=\sqrt{\sum_{k \epsilon K_+}(k-C_f)^2 X(k)}

    Parameters
    ----------
        data : ArrayLike
            Input signal as a 1D array-like.
        sr : int
            Sampling rate of the signal, in Hz.

    Returns
    -------
        float or np.floating
            The standard deviation of the signal.

    Examples
    --------
        >>> import numpy as np
        >>> signal = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        >>> bandwidth(signal, sr=1)
        1.4142135623730951

    References
    ----------
        Klapuri A, Davy M. 2006 Signal processing methods for music transcription.
        New York: Springer. p.136
    """
    _, bandwidth_, _, _ = spectral_moments(data, sr)

    return bandwidth_  # TODO change to variance and let people define bandwidth through percentiles


def skewness(data: ArrayLike, sr: int) -> Union[float, np.floating[Any]]:
    r"""
    Compute the spectral skewness (third spectral moment) of a signal.
    The skewness describes the asymmetry of the spectrum
    around the spectral centroid and is calculated as

        .. math::
            \gamma_1=\frac{\sum_{k \epsilon K_+}(k-C_f)^3 X(k)}{S_f^3}

    Parameters
    ----------
    data : ArrayLike
        Input signal as a 1D array-like.
    sr : int
        Sampling rate of the signal, in Hz.

    Returns
    -------
        float or np.floating
            The skewness of the signal.

    References
    ----------
        Klapuri A, Davy M. 2006 Signal processing methods for music transcription.
        New York: Springer. p.136
    """
    data = check_signal_format(data)
    sr = check_sr_format(sr)

    _, _, skew, _ = spectral_moments(data, sr)
    return skew


def kurtosis(data: ArrayLike, sr: int) -> Union[float, np.floating[Any]]:
    r"""
    Compute the spectral kurtosis (fourth spectral moment) of a signal.
    The skewness describes the 'peakedness' of the spectrum
    and is calculated as

        .. math::
            \gamma_2=\frac{\sum_{k \epsilon K_+}(k-C_f)^4 X(k)}{S_f^4}

    Parameters
    ----------
    data : ArrayLike
        Input signal as a 1D array-like.
    sr : int
        Sampling rate of the signal, in Hz.

    Returns
    -------
        float or np.floating
            The kurtosis of the signal.

    References
    ----------
        Klapuri A, Davy M. 2006 Signal processing methods for music transcription.
        New York: Springer. p.136
    """
    data = check_signal_format(data)
    sr = check_sr_format(sr)

    _, _, _, kurt = spectral_moments(data, sr)
    return kurt


def peak_frequency(data: ArrayLike, sr: int) -> Union[float, np.floating[Any]]:
    """
    Computes the peak frequency of a signal using the Fourier transform.

    The function applies the Fast Fourier Transform (FFT) to the input signal to obtain its frequency spectrum.
    It identifies the frequency corresponding to the maximum magnitude in the spectrum, which represents the dominant
    or peak frequency of the signal.

    Parameters
    ----------
    data : ArrayLike
        1D array-like representing the input signal.
    sampling_rate : float
        Sampling rate of the signal in Hz.

    Returns
    -------
        float
            The peak frequency in Hz.

    Example
    -------
        >>> import numpy as np
        >>> from biosonic.compute.temporal import peak_frequency
        >>> sampling_rate = 1000.0  # 1000 Hz
        >>> t = np.linspace(0, 1.0, int(sampling_rate), endpoint=False)
        >>> signal = np.sin(2 * np.pi * 50 * t)  # 50 Hz sine wave
        >>> freq = peak_frequency(signal, sampling_rate)
        >>> print(freq)
        50.0

    Notes
    -----
        - The function assumes the input signal is real-valued and uniformly sampled.
    """
    data = check_signal_format(data)
    sr = check_sr_format(sr)

    if data.size == 0:
        raise ValueError("Input signal is empty; could not determine peak frequency.")

    freqs, ps = spectrum(data, sr=sr, mode="power")
    assert freqs is not None
    return float(freqs[np.argmax(ps)])


def power_spectral_entropy(
        data: ArrayLike,
        sr: int,
        unit: Literal["bits", "nat", "dits", "bans", "hartleys"] = "bits",
        *args: Any,
        **kwargs: Any
        ) -> Tuple[float, float]:
    """
    Calculates the power spectral entropy as follows:
    1. Compute power spectral density (PSD)
    2. Normalize PSD (interpreted as a probability distribution)
    3. Calculate Shannon-Wiener entropy of normalized PSD

    Args:
        data : ArrayLike
            Input signal as a 1D ArrayLike
        sr : int
            Sampling rate in Hz.
        unit : str, optional
            Desired unit of the entropy, determines the logarithmic base used for calculatein.
            Choose from "bits" (log2), "nat" (ln), or "dits"/"bans"/"hartleys" (log10).
            Defaults to "bits".

    Returns:
        float
            Power spectral entropy.
    References:
        1. https://de.mathworks.com/help/signal/ref/spectralentropy.html accessed January 13th, 2025. 18:34 pm
        2. https://docs.scipy.org/doc/scipy-1.15.2/reference/generated/scipy.stats.entropy.html accessed May 20th 2025, 11:32 am
        3. Shannon C. E. 1948 A mathematical theory of communication. The Bell System Technical Journal XXVII.

    """
    data = check_signal_format(data)
    sr = check_sr_format(sr)

    # _, psd = signal.welch(data, sr, nperseg=N_FFT, noverlap=N_FFT//HOP_OVERLAP) # would return psd - frequency spectrum squared and scaled by sum -
    _, psd = spectrum(data, sr, mode="power")
    psd = exclude_trailing_and_leading_zeros(psd)

    psd_sum: float = np.sum(psd)
    psd_norm = psd / psd_sum
    # Ensure no zero values in normalized power distribution because H is undefined with p=0
    psd_norm = psd_norm[psd_norm > 0]
    return shannon_entropy(psd_norm, unit, *args, **kwargs)


def spectral_features(data: ArrayLike,
                      sr: int,
                      ) -> Dict[str, Union[float, np.floating, NDArray[np.float64]]]:
    """
    Extracts a set of spectral features from a signal.

    Args:
        data : ArrayLike
            The input signal as a 1D ArrayLike.
        sr : int
            Sampling rate of the signal in Hz.

    Retuns:
        dict
        {"mean_frequency" : float,
        "fq_q1": float,
        "fq_median": float,
        "fq_q3": float,
        "spectral_flatness": float,
        "spectral_centroid": float,
        "spectral_skew": float,
        "spectral_kurtosis": float,
        "spectral_sd": float,
        "peak_frequency": float,
        "pse": float
        }
    """
    data = check_signal_format(data)
    check_sr_format(sr)

    fq_q1_bin, fq_median_bin, fq_q3_bin = quartiles(data, sr)
    freqs, ps = spectrum(data, sr, mode="power")

    features = {
        # "mean_frequency" : np.average(freqs, weights=ps),  # because centroid is based on magnitude spectrum
        "fq_q1": fq_q1_bin,
        "fq_median": fq_median_bin,
        "fq_q3": fq_q3_bin,
        "spectral_flatness": flatness(data),
        "spectral_centroid": centroid(data, sr),
        "spectral_sd": bandwidth(data, sr),
        "spectral_skew": skewness(data, sr),
        "spectral_kurtosis": kurtosis(data, sr),
        "peak_frequency": peak_frequency(data, sr),
        "pse": power_spectral_entropy(data, sr)[0]
    }

    return features



import os
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

try:
    import matplotlib.pyplot as plt
    from matplotlib.axes import Axes
    from matplotlib.figure import Figure
except ModuleNotFoundError:
    plt = None  # type: ignore
    Axes = None  # type: ignore
    Figure = None  # type: ignore

import numpy as np
from numpy.typing import ArrayLike
from pandas import DataFrame

from biosonic.compute.spectral import spectrum
from biosonic.compute.spectrotemporal import cepstral_coefficients, cepstrum, spectrogram
from biosonic.compute.utils import check_signal_format, check_sr_format, extract_all_features
from biosonic.filter import mel_filterbank


def plot_spectrogram(
        data: ArrayLike,
        sr: Optional[int] = None,
        db_scale: bool = True,
        cmap: str = 'binary',
        title: Optional[str] = None,
        db_ref: Optional[float] = None,
        dynamic_range: Optional[float] = 100,
        flim: Optional[Tuple[float, float]] = None,
        tlim: Optional[Tuple[float, float]] = None,
        freq_scale: Literal["linear", "log", "mel"] = "linear",
        window_length: int = 512,
        window: Union[str, ArrayLike] = "hann",
        overlap: float = 50,
        noisereduction: Optional[bool] = False,
        n_bands: int = 40,
        corner_frequency: Optional[float] = None,
        after: Optional[Literal["fant", "koenig", "oshaughnessy", "umesh"]] = "oshaughnessy",
        plot: Optional[Any] = None,
        show_amplitude_bar: Optional[bool] = True,
        **kwargs: Any
    ) -> Any:
    """
    Plot a time-frequency spectrogram with optional dB scaling and frequency axis transformations.

    Parameters
    ----------
    data : Union[np.ndarray, Tuple[np.ndarray, np.ndarray, np.ndarray]],
        Input signal. Either as precomputed spectrogram (S, t, f) or 1D signal array.
    sr : int
        Sampling rate in Hz.
    db_scale : bool, optional
        Whether to convert the spectrogram to decibel scale. Default is True.
    cmap : str, optional
        Colormap for the spectrogram.
    title : str, optional
        Title of the plot. If None, title is generated based on 'freq_scale'.
    db_ref : float or None
        Reference for dB scaling (max if None).
    dynamic_range : float or None, optional
        Clip values below (max_dB - dynamic_range) after dB conversion. Set to None if no clipping is desired.
        Useful for suppressing low-amplitude noise. Default is 100.
    flim : tuple of float or None
        Frequency limits in Hz.
    tlim : tuple of float or None
        Time limits in seconds.
    freq_scale : {"linear", "log", "mel"}, optional
        Frequency axis scale. Choose "linear", "log", or "mel". Default is "linear".
    window_length : int, optional
        Length of the window for STFT (in samples). Default is 512.
    window : str or ArrayLike, optional
        Windowing function used for the STFT. Default is "hann".
    zero_padding : int, optional
        Number of zeros to pad each windowed frame. Default is 0.
    overlap : float, optional
        Percentage of overlap between successive windows (0 to 100). Default is 50.
    noisereduction : bool, optional
        Whether to apply noise reduction. Default is False (no reduction).
    n_bands : int, optional
        Number of frequency bands for mel scaling. Default is 40.
    corner_frequency : float, optional
        Corner frequency for perceptual frequency scaling (used in mel scale).
    after (Literal): Choice of Mel scale formula to use.
            - 'fant': Classic formula with `a=b=1000`: `F_m = a * np.log(1 + f / b)`
            - 'koenig': (Not yet implemented)
            - 'oshaughnessy' or 'beranek': Commonly used formula as fant, but with `a=2595`, `b=700`
            - 'umesh': Formula using rational function with `a=0.0004`, `b=0.603`.
    plot : tuple(matplotlib.figure.Figure, matplotlib.axes.Axes), optional
        Existing matplotlib Figure and Axes objects to plot into. If None, a new figure and axes are created.
    **kwargs : dict
        Additional keyword arguments passed to `matplotlib.pyplot.imshow`.
    """
    if plt is None:
        raise ImportError("matplotlib is required for plotting. Install it with: pip install biosonic[plot]")

    # Precomputed spectrogram
    if isinstance(data, tuple) and len(data) == 3:
        Sx, t, f = data

    # Raw signal + sr
    elif isinstance(data, np.ndarray):
        if sr is None:
            raise ValueError("sr must be provided when passing a signal array.")
        Sx, t, f = spectrogram(
            data=data,
            sr=sr,
            window_length=window_length,
            window=window,
            overlap=overlap,
            noisereduction=noisereduction,
            complex_output=False
        )

    else:
        raise TypeError("data must be either a (S, t, f) tuple or a 1D np.ndarray signal")

    if freq_scale == "mel":
        if sr is None:
            raise ValueError("Sample rate must be provided for mel frequency scale.")

        fmin = flim[0] if flim else 0.0
        fmax = flim[1] if flim and flim[1] else sr / 2

        fb, f_centers = mel_filterbank(n_bands, window_length, sr, fmin=fmin, fmax=fmax, corner_frequency=corner_frequency, after=after)
        f = f_centers
        # Sx : np.ndarray = np.einsum("...ft,mf->...mt", Sx, fb, optimize=True)
        Sx = fb @ Sx

    # Apply dB scale
    if db_scale:
        ref = np.max(Sx) if db_ref is None else db_ref
        Sx = 20 * np.log10(Sx / ref + 1e-30)  # add small value to avoid log(0)

        if dynamic_range is not None:
            if dynamic_range is not None:
                Sx = np.maximum(Sx, -dynamic_range)

    # Apply frequency limits
    if flim is not None and freq_scale in ("linear", "log"):  # already handled in mel case
        fmin, fmax = flim
        mask = (f >= fmin) & (f <= fmax)
        f = f[mask]
        Sx = Sx[mask, :]

    # Apply time limits
    if tlim is not None:
        tmin, tmax = tlim
        mask = (t >= tmin) & (t <= tmax)
        t = t[mask]
        Sx = Sx[:, mask]

    # Plot
    if plot is not None:
        fig, ax = plot
    else:
        fig, ax = plt.subplots(figsize=(10, 5))

    extent = (
        float(t[0]),
        float(t[-1]),
        float(f[0]),
        float(f[-1]),
    )

    im = ax.imshow(Sx, aspect='auto', origin='lower', extent=extent, cmap=cmap, **kwargs)
    ax.set_ylabel("Frequency [Hz]")
    ax.set_xlabel("Time [s]")

    if freq_scale == "log":
        ax.set_yscale("log")
        ax.set_ylim(float(min(f[f > 0])), float(f[-1]))

    if title is None:
        title = f"{freq_scale}-scaled spectrogram"
    ax.set_title(title)

    if show_amplitude_bar:
        fig.colorbar(im, ax=ax, label=("Amplitude [dB]" if db_scale else "Magnitude"))
    plt.tight_layout()

    if plot is None:
        plt.show()

    return fig, ax


def plot_cepstrum(
        data: ArrayLike,
        sr: int,
        min_quefrency: Optional[float] = None,
        max_quefrency: Optional[float] = None,
        log_scale: bool = False,
        ylim: Optional[Tuple[float, float]] = None,
        title: Optional[str] = None,
        **kwargs: Any
) -> None:
    """
    Plot the cepstrum against quefrency (in seconds).

    Parameters
    ----------
    data : ArrayLike
        Signal to use for cepstrum calculation.
    sr : int
        Sampling rate of the original signal.
    max_quefrency : float, optional
        Maximum quefrency (in seconds) to plot. Defaults to 0.05s.
    log_scale : bool, optional
        Wether to log-scale the y-axis. Defaults to False.
    ylim : tuple, optional
        Limits for the y-axis.
    title : str or None
        Title for the plot. If None, a default will be generated.
    """
    if plt is None:
        raise ImportError("matplotlib is required for plotting. Install it with: pip install biosonic[plot]")

    ceps, quefs = cepstrum(data, sr, **kwargs)

    if max_quefrency is None:
        max_quefrency = len(data) / sr

    if min_quefrency is None:
        min_quefrency = 0

    mask = (quefs >= min_quefrency) & (quefs <= max_quefrency)

    plt.plot(quefs[mask], ceps[mask], color='steelblue')
    plt.xlabel("Quefrency (s)")
    if log_scale:
        plt.yscale("log")
    if ylim:
        plt.ylim(ylim)
    plt.ylabel("Amplitude")
    plt.title(title or f"Cepstrum (Sampling rate: {sr} Hz)")
    plt.grid(True)
    plt.tight_layout()
    plt.show()


def plot_cepstral_coefficients(
        data: ArrayLike,
        sr: int,
        window_length: int,
        n_filters: int = 32,
        n_ceps: int = 40,
        pre_emphasis: float = 0.97,
        fmin: float = 0.0,
        fmax: Optional[float] = None,
        filterbank_type: Literal["mel", "linear", "log"] = "mel",
        cmap: Optional[str] = "grey",
        **kwargs: Any
    ) -> None:
    """
    Compute and plot cepstral coefficients over time from audio data.

    This function calculates cepstral coefficients using a specified filterbank
    type and visualizes them with time on the x-axis and
    cepstral coefficient indices on the y-axis.

    Parameters
    ----------
    data : ArrayLike
        Input audio signal (1D array-like).
    sr : int
        Sampling rate of the audio signal in Hz.
    window_length : int
        Length of the analysis window in samples.
    n_filters : int, optional
        Number of filters in the filterbank. Default is 32.
    n_ceps : int, optional
        Number of cepstral coefficients to compute and plot. Default is 40.
    pre_emphasis : float, optional
        Pre-emphasis filter coefficient. Default is 0.97.
    fmin : float, optional
        Minimum frequency for the filterbank in Hz. Default is 0.0.
    fmax : float, optional
        Maximum frequency for the filterbank in Hz. Defaults to Nyquist frequency if None.
    filterbank_type : {'mel', 'linear', 'log'}, optional
        Type of filterbank to use for cepstral coefficient calculation.
        Default is 'mel'.
    cmap : str or None, optional
        Matplotlib colormap name for plotting. Default is 'grey'.
    **kwargs : dict, optional
        Additional keyword arguments passed to the cepstral_coefficients function.
    """
    if plt is None:
        raise ImportError("matplotlib is required for plotting. Install it with: pip install biosonic[plot]")

    ceps = cepstral_coefficients(
        data,
        sr,
        window_length,
        n_filters,
        n_ceps,
        pre_emphasis=pre_emphasis,
        fmin=fmin,
        fmax=fmax,
        filterbank_type=filterbank_type,
        **kwargs)

    times = np.linspace(0, len(data) / sr, ceps.shape[0])
    plt.xlabel("Time [s]")
    plt.ylabel("Cepstral Coefficient Index")
    plt.imshow(ceps, origin="lower", aspect="auto", extent=(times[0], times[-1], 0, n_ceps), cmap=cmap)


def plot_features(
        data: ArrayLike,
        sr: int,
        features: Optional[dict[str, Any]] = None,
        spec_kwargs: Optional[dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
    """
    Plot audio signal features using precomputed feature dictionary.

    Parameters
    ----------
    data : ArrayLike
        Audio time series data.
    sr : int
        Sampling rate of the audio data in Hz.
     **kwargs : dict[str, Any]
            Optional parameters for dominant frequency estimation.
    """
    if plt is None:
        raise ImportError("matplotlib is required for plotting. Install it with: pip install biosonic[plot]")

    data = check_signal_format(data)
    sr = check_sr_format(sr)

    if not features:
        features = extract_all_features(data, sr, **kwargs)

    _, times, _ = spectrogram(data, sr)
    freq_ms, ms = spectrum(data, sr)

    dom_freqs = features["dominant_freqs"]
    all_candidates = [[(float(f), 1.0) if f > 0 else (0.0, 0.0)] for f in dom_freqs]

    # Spectrogram with Dominant Frequencies
    fig = plt.figure(figsize=(12, 12))
    ax1 = fig.add_subplot(3, 1, 1)

    if spec_kwargs is None:
        spec_kwargs = {}

    plot_pitch_on_spectrogram(
        data=data,
        sr=sr,
        time_points=times,
        all_candidates=all_candidates,
        show_strongest=True,
        db_scale=True,
        title="Spectrogram with Dominant Frequencies",
        cmap="binary",
        plot=(fig, ax1),
        **spec_kwargs
    )

    # Spectrum
    ax2 = fig.add_subplot(3, 1, 2)
    ax2.set_title("Magnitude Spectrum with Spectral Features")
    if not isinstance(freq_ms, np.ndarray):
        raise TypeError("Expected 'freqs_ps' to be an ndarray.")
    if not isinstance(freq_ms, np.ndarray):
        raise TypeError("Expected 'freqs_ps' to be an ndarray.")
    cutoff = len(freq_ms) // 3
    ax2.plot(freq_ms[:-cutoff], ms[:-cutoff], label="Magnitude Spectrum", color="#A2A2A2")

    ax2.axvline(features["peak_frequency"], color="#7951A2C6", linestyle="-", label="Peak Frequency (kHz)")
    ax2.axvline(features["fq_median"], color="#48ad46b5", linestyle="-", label="Median")
    ax2.axvline(features["fq_q1"], color="#88d253aa", linestyle="-", label="Q1")
    ax2.axvline(features["fq_q3"], color="#267746A9", linestyle="-", label="Q3")
    ax2.fill_betweenx(
        y=[0, max(ms)],
        x1=features["spectral_centroid"] - features["spectral_sd"],
        x2=features["spectral_centroid"] + features["spectral_sd"],
        color='grey',
        alpha=0.2,
        label='Bandwidth'
    )
    ax2.axvline(features["spectral_centroid"], color="#00BFFF7E", linestyle="-", label="Centroid")
    ax2.set_xlabel("Frequency [Hz]")
    ax2.set_ylabel("Magnitude")
    ax2.legend(loc="upper left", bbox_to_anchor=(1.02, 1), borderaxespad=0)

    # Waveform
    ax3 = fig.add_subplot(3, 1, 3)
    ax3.set_title("Waveform with Amplitude Envelope and Time-domain Features")
    times_waveform = np.linspace(0, len(data) / sr, num=len(data))
    ax3.plot(times_waveform, data, label="Waveform", color="grey", alpha=0.3)
    ax3.plot(times_waveform, features["amplitude_envelope"], label="Amplitude Envelope", color="#A2A2A2")
    ax3.axvline(features["t_median"], color="#48ad46b5", linestyle="-", label="Median")
    ax3.axvline(features["t_q1"], color="#88d253aa", linestyle="-", label="Q1")
    ax3.axvline(features["t_q3"], color="#267746A9", linestyle="-", label="Q3")
    ax3.fill_betweenx(
        y=[0, np.max(features["amplitude_envelope"])],
        x1=features["temporal_centroid"] - features["temporal_sd"],
        x2=features["temporal_centroid"] + features["temporal_sd"],
        color='grey',
        alpha=0.2,
        label='Bandwidth'
    )
    ax3.axvline(features["temporal_centroid"], color="#00BFFF7E", linestyle="-", label="Centroid")
    ax3.set_xlabel("Time [s]")
    ax3.set_ylabel("Amplitude")
    ax3.legend(loc="upper left", bbox_to_anchor=(1.02, 1), borderaxespad=0)

    plt.tight_layout()
    plt.show()


def plot_pitch_candidates(
        time_points: ArrayLike,
        all_candidates: ArrayLike,
        show_strongest: bool = True,
        tlim: Optional[Tuple[float, float]] = None,
        ax: Optional[Axes] = None
    ) -> Optional[Axes]:
    """
    Plot pitch candidates over time.

    Parameters
    ----------
    time_points : list of float
        Time stamps for each frame.
    all_candidates : list of list of tuple(float, float)
        List containing, for each frame, a list of (pitch, strength) tuples.
    show_strongest : bool
        If True, highlight the strongest voiced candidate per frame.
    """
    if plt is None:
        raise ImportError("matplotlib is required for plotting. Install it with: pip install biosonic[plot]")

    if ax is None:
        fig, ax = plt.subplots(figsize=(10, 5))

    if all(isinstance(p, (int, float, np.number)) for p in all_candidates):
        all_candidates = [[(p, 1.0)] for p in all_candidates]

    # Plot all candidates
    for t, candidates in zip(time_points, all_candidates):
        if tlim and not (tlim[0] <= t <= tlim[1]):
            continue
        for pitch, _ in candidates:
            if pitch > 0:
                ax.plot(t, pitch, 'k.', alpha=0.3)

    # Optionally plot the strongest voiced candidate
    if show_strongest:
        times = []
        pitches = []
        for t, candidates in zip(time_points, all_candidates):
            if tlim and not (tlim[0] <= t <= tlim[1]):
                continue
            voiced = [c for c in candidates if c[0] > 0]
            if voiced:
                best = max(voiced, key=lambda x: x[1])
                times.append(t)
                pitches.append(best[0])
        ax.scatter(times, pitches, color=(0.7, 0.1, 0.1, 0.3), marker="o", label='Strongest pitch candidate')

    if tlim:
        ax.set_xlim(tlim)

    if ax is None:
        plt.title("Autocorrelation based pitch tracking")
        plt.xlabel("Time [s]")
        plt.ylabel("Pitch [Hz]")
        plt.grid(True)
        plt.legend()
        plt.tight_layout()
        plt.show()
        return None

    return ax


def plot_pitch_on_spectrogram(
    data: ArrayLike,
    sr: int,
    time_points: ArrayLike,
    all_candidates: ArrayLike,
    window_length: int = 512,
    overlap: int = 50,
    show_strongest: bool = True,
    db_scale: bool = True,
    flim: Optional[Tuple[float, float]] = None,
    tlim: Optional[Tuple[float, float]] = None,
    title: str = "Spectrogram with Pitch Candidates",
    cmap: str = "binary",
    plot: Optional[Tuple[Figure, Axes]] = None,
    **kwargs: Any
) -> None:
    """
    Plot a spectrogram of the input audio data and overlay pitch candidates.

    This function computes and displays a spectrogram of the given audio data,
    then overlays pitch candidates over time. It can optionally highlight the
    strongest pitch candidate per time frame.

    Parameters
    ----------
    data : ArrayLike
        Audio time series data.
    sr : int
        Sampling rate of the audio data in Hz.
    time_points : ArrayLike
        Time stamps corresponding to each frame of pitch candidates.
    all_candidates : ArrayLike
        List or array of pitch candidate tuples (pitch, strength) for each time frame.
    window_length : int, optional
        Window length (in samples) for the spectrogram. Default is 512.
    overlap : int, optional
        Overlap between windows (in samples) for the spectrogram. Default is 50.
    show_strongest : bool, optional
        If True, highlights the strongest voiced pitch candidate per frame. Default is True.
    db_scale : bool, optional
        Whether to display the spectrogram in decibel scale. Default is True.
    flim : tuple of float, optional
        Frequency limits (min_freq, max_freq) to display in the spectrogram. Default is None (no limit).
    tlim : tuple of float, optional
        Time limits (start_time, end_time) for the plot. Default is None (full duration).
    title : str, optional
        Title of the plot. Default is "Spectrogram with Pitch Candidates".
    cmap : str, optional
        Colormap to use for the spectrogram. Default is 'binary'.
    plot : tuple of (Figure, Axes), optional
        Existing matplotlib Figure and Axes to plot on. If None, a new figure is created.
    """
    if plt is None:
        raise ImportError("matplotlib is required for plotting. Install it with: pip install biosonic[plot]")

    if plot is None:
        fig, ax = plt.subplots(figsize=(10, 5))
    else:
        fig, ax = plot

    plot_spectrogram(
        data,
        sr,
        overlap=overlap,
        db_scale=db_scale,
        cmap=cmap,
        flim=flim,
        tlim=tlim,
        title=title,
        window_length=window_length,
        plot=(fig, ax),
        **kwargs
    )

    plot_pitch_candidates(
        time_points=time_points,
        all_candidates=all_candidates,
        show_strongest=show_strongest,
        tlim=tlim,
        ax=ax,
        )

    if plot is None:
        plt.show()


def plot_boundaries_on_spectrogram(
    data: ArrayLike,
    sr: int,
    segments: List[Dict[str, float]],
    **kwargs: Any
    ) -> None:
    """
    Plot a spectrogram of the input audio data and overlay vertical lines indicating segment boundaries.

    Parameters
    ----------
    data : ArrayLike
        Audio time series data.
    sr : int
        Sampling rate of the audio data.
    segments : List[Dict[str, float]]
        A list of segment boundary dictionaries. Each dictionary should
        contain keys "begin" and "end", representing the start and end
        times (in seconds or frames, depending on the spectrogram scale).
    **kwargs : Any
        Additional keyword arguments passed to the spectrogram plotting function.
    """
    if plt is None:
        raise ImportError("matplotlib is required for plotting. Install it with: pip install biosonic[plot]")

    fig, ax = plt.subplots()

    plot_spectrogram(data, sr, plot=(fig, ax), **kwargs)

    tlim = kwargs.get("tlim", None)

    for segment in segments:
        if tlim is None or (segment["begin"] >= tlim[0] and segment["end"] <= tlim[1]):
            ax.axvline(segment["begin"], color="#95c5478d", linestyle="-")
            ax.axvline(segment["end"], color="#ffa2007f", linestyle="-")

        plt.legend(["Segment Start", "Segment End"])

    plt.show()


def plot_spectrogram_catalogue(
    df: DataFrame,
    column: str = "waveform",
    per_page: int = 20,
    ncols: int = 2,
    save_dir: Optional[str] = None,
    show: bool = True,
    prefix: str = "spectrograms",
    title_columns: List[str] = ["filename"],
    **kwargs: Any
) -> None:
    """
    Plot a catalogue of spectrograms from a DataFrame.

    This function generates spectrogram plots for waveforms stored in a DataFrame.
    The plots are arranged in a paginated grid and can be saved to disk.

    Parameters
    ----------
    df : pandas.DataFrame
        Input DataFrame containing waveform data and associated metadata.
    column : str, default="waveform"
        Column name in `df` that contains the waveform arrays.
    per_page : int, default=20
        Number of spectrograms to display per page.
    ncols : int, default=2
        Number of subplot columns per page.
    save_dir : str or None, optional
        Directory to save the generated figure pages. If None, figures are not saved.
    show : bool, default=True
        If True, display the figures interactively. If False, figures are closed after saving.
    prefix : str, default="spectrograms"
        Prefix for saved figure filenames if `save_dir` is provided.
    title_columns : list of str, default=["filename"]
        List of column names from `df` to include in subplot titles, joined by a vertical bar (`|`).
    **kwargs : dict, optional
        Additional keyword arguments passed to the underlying `plot_spectrogram` function.

    Returns
    -------
    None
        This function does not return anything. It produces matplotlib figures,
        optionally displaying them or saving them to disk.

    Notes
    -----
    - The function expects each row of the DataFrame to contain:
        * `df[column]`: waveform array-like
        * `df["sr"]`: sampling rate (int)
    - If a waveform entry is `None`, its subplot will be empty.

    Examples
    --------
    >>> import pandas as pd
    >>> data = {
    ...     "ID": [1, 2],
    ...     "waveform": [wave1, wave2],
    ...     "sr": [16000, 16000]
    ... }
    >>> df = pd.DataFrame(data)
    >>> plot_spectrogram_catalogue(df, per_page=2, ncols=2, show=False, save_dir="plots")
    """
    if plt is None:
        raise ImportError("matplotlib is required for plotting. Install it with: pip install biosonic[plot]")

    if save_dir is not None:
        os.makedirs(save_dir, exist_ok=True)

    n = len(df)
    n_pages = int(np.ceil(n / per_page))

    for page in range(n_pages):
        start = page * per_page
        end = min((page + 1) * per_page, n)
        subset = df.iloc[start:end]

        nrows = int(np.ceil(len(subset) / ncols))
        fig, axes = plt.subplots(nrows, ncols, figsize=(5*ncols, 2*nrows))
        axes = np.atleast_1d(axes).reshape(nrows, ncols)

        for idx, (_, row) in enumerate(subset.iterrows()):
            r, c = divmod(idx, ncols)
            ax_spec = axes[r, c]

            snippet = row[column]
            if snippet is None:
                ax_spec.axis("off")
                continue

            sr = row["sr"]

            plot_spectrogram(
                row["waveform"],
                sr,
                title="|".join(str(row[col]) for col in title_columns),
                plot=(fig, ax_spec),
                show_amplitude_bar=False,
                **kwargs
            )

        plt.tight_layout()

        if save_dir is not None:
            filename = os.path.join(save_dir, f"{prefix}_{page+1:03d}.png")
            fig.savefig(filename, dpi=150)
            print(f"Saved {filename}")

        if show:
            plt.show()
        else:
            plt.close(fig)



from typing import Any, List, Optional, Tuple

import numpy as np
import scipy.optimize
from numpy.typing import ArrayLike, NDArray
from scipy.fft import irfft, rfft
from scipy.signal import windows

from biosonic.compute.utils import frame_signal


def _difference_function(x: ArrayLike, max_lag: int) -> ArrayLike:
    """YIN difference function d(τ) = sum_j (x_j - x_{j+τ})^2"""
    N = len(x)
    d = np.zeros(max_lag + 1)
    for lag in range(1, max_lag + 1):
        d[lag] = np.sum((x[:N - lag] - x[lag:]) ** 2)
    return d


def _cumulative_mean_normalized_difference(d: ArrayLike) -> ArrayLike:
    """CMND function from the difference function."""
    cmndf = np.zeros_like(d)
    cmndf[0] = 1  # Avoid divide-by-zero
    cumsum = np.cumsum(d[1:])
    cmndf[1:] = d[1:] * np.arange(1, len(d)) / cumsum
    return cmndf


def _parabolic_interpolation(cmndf: ArrayLike, tau: int) -> float:
    """Refine τ estimate using parabolic interpolation."""
    if tau <= 0 or tau >= len(cmndf) - 1:
        return float(tau)
    alpha = cmndf[tau - 1]
    beta = cmndf[tau]
    gamma = cmndf[tau + 1]
    denominator = alpha + gamma - 2 * beta
    if denominator == 0:
        return float(tau)
    shift = 0.5 * (alpha - gamma) / denominator
    return float(tau + shift)


def _yin_single_window(
    x: ArrayLike,
    sr: int,
    window_length: int,
    time_step: int,
    bounds: Tuple[int, int],
    threshold: float = 0.1
) -> Optional[float]:
    """Estimate fundamental frequency from a single frame using YIN."""
    min_lag, max_lag = bounds
    frame = x[time_step:time_step + window_length]
    if len(frame) < window_length:
        return None

    # remove DC offset
    frame = frame - np.mean(frame)

    d = _difference_function(frame, max_lag)
    cmndf = _cumulative_mean_normalized_difference(d)

    max_lag = min(max_lag, len(cmndf) - 1)
    for i in range(min_lag, max_lag):
        if cmndf[i] < threshold:
            refined_tau = _parabolic_interpolation(cmndf, i)
            return sr / refined_tau

    return None


def yin(
    data: ArrayLike,
    sr: int,
    window_length: int,
    time_step_sec: float,
    flim: Tuple[int, int],
    threshold: float = 0.1
) -> Tuple[ArrayLike, ArrayLike]:
    """YIN pitch tracking over an entire signal. **Not yet finished**

    Returns
    -------
        times : ArrayLike
            time points (in seconds)
        frequencies : ArrayLike
            estimated f₀ at each time point
    """
    step_size = int(time_step_sec * sr)
    num_steps = (len(data) - window_length) // step_size

    min_lag = int(sr / flim[0])
    max_lag = int(sr / flim[1])

    times = []
    frequencies = []

    for i in range(num_steps):
        time_step = i * step_size
        time_sec = time_step / sr
        f0 = _yin_single_window(
            data,
            sr,
            window_length,
            time_step,
            bounds=(min_lag, max_lag),
            threshold=threshold
        )
        if f0 is not None:
            times.append(time_sec)
            frequencies.append(f0)

    return np.array(times), np.array(frequencies)


def _preprocess_for_pitch_(
        data: ArrayLike,
        sr: int
    ) -> ArrayLike:
    """
    Soft upsampling via frequency filtering and iFFT with longer FFT size
    to remove sidelobe of the FT of the Hanning window near f_nyquist as described in [1].

    References:
    ----------
    1. Boersma P. 1993 Accurate short-term analysis of the fundamental
    frequency and the harmonics-to-noise ratio of a sampled sound.
    IFA Proceedings 17, 97–110.
    """
    spectrum = rfft(data)
    nyquist = sr / 2
    freqs = np.linspace(0, nyquist, len(spectrum))

    # Soft taper from 95% to 100% Nyquist
    taper = np.ones_like(freqs)
    taper_start = 0.95 * nyquist
    taper_end = nyquist
    taper_region = (freqs >= taper_start) & (freqs <= taper_end)
    taper[taper_region] = 0.5 * (1 + np.cos(np.pi * (freqs[taper_region] - taper_start) / (taper_end - taper_start)))
    taper[freqs > nyquist] = 0

    spectrum *= taper

    # inverse FFT back to time domain
    filtered_signal = irfft(spectrum, n=len(data))

    return filtered_signal


def _sinc_interpolation(
        y: NDArray[np.floating],
        x: float,
        max_depth: int
        ) -> float:
    midleft = np.floor(x).astype(int)
    midright = midleft + 1
    if x == midleft:
        return np.float64(y[midleft])

    max_depth = min(min(max_depth, midleft), len(y) - midright)
    assert max_depth > 2

    left = midright - max_depth
    right = midleft + max_depth

    result = 0.0

    a = np.pi * (x - midleft)
    halfsina = 0.5 * np.sin(a)
    aa = a / (x - left + 1)
    daa = np.pi / (x - left + 1)
    for i in range(midleft, left - 1, -1):
        d = halfsina / a * (1 + np.cos(aa))
        result += y[i] * d
        a += np.pi
        aa += daa
        halfsina = -halfsina

    a = np.pi * (midright - x)
    halfsina = 0.5 * np.sin(a)
    aa = a / (right - x + 1)
    daa = np.pi / (right - x + 1)
    for i in range(midright, right + 1):
        d = halfsina / a * (1 + np.cos(aa))
        result += y[i] * d
        a += np.pi
        aa += daa
        halfsina = -halfsina

    return result


def _improve_sinc_maximum(y: ArrayLike, x: float, max_depth: int) -> Tuple[float, float]:
    """
    Refine the maximum using sinc interpolation and Brent's method.
    Returns (refined_x, value_at_refined_x)
    """
    assert 0 <= x < len(y)

    def _neg_sinc_interp(x_: float) -> float:
        return -_sinc_interpolation(y, x_, max_depth)

    xmin, fval, _, _ = scipy.optimize.brent(
        _neg_sinc_interp,
        brack=(x - 1, x, x + 1),
        tol=1e-10,
        full_output=True,
        maxiter=60
    )
    return xmin, -fval


def _find_pitch_candidates_(
        ac: ArrayLike,
        sr: int,
        min_pitch: int,
        max_pitch: int,
        num_candidates: int = 4,
        octave_cost: float = 0.01
        ) -> ArrayLike:
    """
    Find pitch candidates based on autocorrelation peaks.
    """
    min_lag = int(sr / max_pitch)
    max_lag = int(sr / min_pitch)

    candidates: list[Tuple[float, float]] = []

    max_depth = 8  # window for sinc interpolation, can be tuned
    for lag in range(min_lag + 1, max_lag - 1):
        if ac[lag] > ac[lag - 1] and ac[lag] > ac[lag + 1]:
            # Use sinc interpolation for sub-sample lag refinement
            refined_lag, _ = _improve_sinc_maximum(ac, float(lag), max_depth)

            # cost function (Boersma 1993, eq 26)
            r_tau = _sinc_interpolation(ac, refined_lag, max_depth)
            strength = r_tau - octave_cost * 2 * np.log(min_pitch * refined_lag)

            # convert to pitch
            pitch = sr / refined_lag if refined_lag != 0 else 0
            candidates.append((pitch, strength))

    # Sort by strength and take top N-1
    candidates = sorted(candidates, key=lambda x: -x[1])[:num_candidates - 1]

    # normalize to [0,1]
    max_strength = max(s for _, s in candidates) if candidates else 1.0
    if max_strength > 0:
        candidates = [(p, s / max_strength) for p, s in candidates]
    else:
        candidates = [(p, 0.0) for p, s in candidates]

    return candidates


def _transition_cost(
        F1: float,
        F2: float,
        voiced_unvoiced_cost: float,
        octave_jump_cost: float
    ) -> float:
    if F1 == 0.0 and F2 == 0.0:
        return 0.0
    elif F1 == 0.0 or F2 == 0.0:
        return voiced_unvoiced_cost
    else:
        return float(octave_jump_cost * abs(np.log2(F1 / F2)))


def _viterbi_pitch_path(
        all_candidates: List[List[Tuple[float, float]]],
        voiced_unvoiced_cost: float = 0.2,
        octave_jump_cost: float = 0.2
    ) -> List[float]:
    """
    Finds the globally optimal pitch path using dynamic programming.

    Parameters
    ----------
    all_candidates: List of lists of (pitch in Hz, strength)
    voiced_unvoiced_cost: Cost of voiced/unvoiced transition
    octave_jump_cost: Cost of pitch discontinuity in octaves

    Returns
    -------
    path
        List of chosen pitch values, one per frame
    """
    num_frames = len(all_candidates)
    path_costs = []
    back_pointers: List[List[Any]] = []

    # initialization
    prev_costs = [-strength for _, strength in all_candidates[0]]
    path_costs.append(prev_costs)
    back_pointers.append([None] * len(all_candidates[0]))

    # dynamic programming
    for t in range(1, num_frames):
        frame_costs = []
        frame_back_ptrs = []
        for j, (curr_pitch, curr_strength) in enumerate(all_candidates[t]):
            best_cost = float('inf')
            best_prev_idx = None
            for i, (prev_pitch, _) in enumerate(all_candidates[t-1]):
                trans_cost = _transition_cost(
                    prev_pitch, curr_pitch,
                    voiced_unvoiced_cost, octave_jump_cost
                )
                total_cost = path_costs[t-1][i] + trans_cost - curr_strength
                if total_cost < best_cost:
                    best_cost = total_cost
                    best_prev_idx = i
            frame_costs.append(best_cost)
            frame_back_ptrs.append(best_prev_idx)
        path_costs.append(frame_costs)
        back_pointers.append(frame_back_ptrs)

    # backtrace
    path = [0.0] * num_frames
    idx = int(np.argmin(path_costs[-1]))
    for t in reversed(range(num_frames)):
        pitch, _ = all_candidates[t][idx]
        path[t] = pitch
        idx = back_pointers[t][idx] if back_pointers[t][idx] is not None else 0

    return path


# def hnr() -> float:
#     """
#     Calculate harmonics-to-noise ratio as in Boersma 1993. Returns value in dB.
#     """
#     r_tmax =
#     return 10 * np.log10(r_tmax/1-r_tmax)


def _autocorr(
        frame: ArrayLike,
        pad_width_for_pow2: int
    ) -> NDArray:
    # 3.5 and 3.6 append half a window length of zeroes
    # plus enough until the length is a power of two
    frame = np.pad(frame, (0, pad_width_for_pow2), mode='constant', constant_values=0)

    # 3.7 perform fft
    spec = rfft(frame)

    # 3.8 square samples in frequency domain
    power_spec = spec * np.conj(spec)

    # 3.9 ifft of power spectrum
    lag_domain = irfft(power_spec)

    return lag_domain[: len(frame) // 2]


def boersma(
        data: ArrayLike,
        sr: int,
        min_pitch: int = 75,
        max_pitch: int = 600,
        timestep: float = 0.01,
        silence_thresh: float = 0.03,
        voicing_thresh: float = 0.45,
        max_candidates: int = 15,
        octave_cost: float = 0.01,
        octave_jump_cost: float = 0.35,
        voiced_unvoiced_cost: float = 0.14,
        plot: bool = False,
        **kwargs: Any
    ) -> Tuple[ArrayLike, ArrayLike, List[List[Tuple[float, float]]], ArrayLike]:
    """
    Estimate the fundamental frequency track of a signal using a Praat/Boersma-style autocorrelation method.

    Parameters
    ----------
    data : ArrayLike
        Input audio signal (mono, normalized to [-1, 1]).
    sr : int
        Sampling rate (Hz).
    min_pitch : int, optional
        Minimum pitch to search for (Hz). Default is 75.
    max_pitch : int, optional
        Maximum pitch to search for (Hz). Default is 600.
    timestep : float, optional
        Time step between pitch frames (seconds). Default is 0.01.
    silence_thresh : float, optional
        Silence threshold for voicing decision. Default is 0.03.
    voicing_thresh : float, optional
        Voicing threshold for candidate selection. Default is 0.45.
    max_candidates : int, optional
        Maximum number of pitch candidates per frame. Default is 15.
    octave_cost : float, optional
        Cost for octave errors in candidate selection. Default is 0.01.
    octave_jump_cost : float, optional
        Cost for octave jumps in dynamic programming. Default is 0.35.
    voiced_unvoiced_cost : float, optional
        Cost for voiced/unvoiced transitions in dynamic programming. Default is 0.14.
    plot : bool, optional
        If True, plot the pitch track on a spectrogram. Default is False.
    **kwargs : Any
        Additional keyword arguments for plotting.

    Returns
    -------
    time_points : np.ndarray
        Array of time points (seconds) for each pitch frame.
    pitch_track : np.ndarray
        Array of estimated F0 values (Hz) for the best path.
    all_candidates : list of list of (float, float)
        List of candidate (frequency, strength) tuples for each frame.
    intensities : np.ndarray
        Array of frame intensities (relative to global peak).

    References
    ----------
    1. Boersma P. 1993 Accurate short-term analysis of the fundamental
       frequency and the harmonics-to-noise ratio of a sampled sound.
       IFA Proceedings 17, 97–110.
    2. Anikin A. 2019. Soundgen: an open-source tool for synthesizing
       nonverbal vocalizations. Behavior Research Methods, 51(2), 778-792.
    """

    # make sure, the signal is inside the bounds [-1, 1]
    if np.max(np.abs(data)) > 1:
        raise ValueError("the signal needs to be within the bounds [-1, 1]")

    if min_pitch >= max_pitch or max_pitch >= sr / 2:
        raise ValueError("max_pitch should be greater than min_pitch and below the nyquist frequency.")

    # not enough resolution above half the niquist frequency
    # -> amend pitch ceiling if applicable. From Soundgen (see references)
    # max_pitch = min(max_pitch, sr / 4)

    window_length = 3 * (1 / min_pitch)  # three periods of minimum frequency
    data_preprocessed = data  # _preprocess_for_pitch_(data, sr)
    global_peak: float = np.max(np.abs(data_preprocessed - np.mean(data_preprocessed)))

    # global_peak: float = np.max(np.abs(data_preprocessed))
    window_length_samples = int(window_length * sr)

    # precalculate for padding to power of two (step 3.6)
    # - I do this here to save computation time despite it being a bit less readable
    n = window_length_samples + np.floor(window_length_samples/2)
    next_pow2 = 2 ** np.ceil(np.log2(n)).astype(int)
    pad_width_for_pow2 = next_pow2 - window_length_samples  # full pad length needed including half a window size

    # 1. windowing
    framed_signal = frame_signal(data_preprocessed, sr, window_length_samples, timestep, normalize=False)
    window = windows.get_window("hann", window_length_samples)
    autocorr_hann = _autocorr(window, pad_width_for_pow2)
    autocorr_hann /= np.max(autocorr_hann)
    all_candidates = []
    intensities = []

    for frame in framed_signal:
        # 3.2 subtract local average
        frame = frame - np.mean(frame)
        local_peak: float = np.max(np.abs(frame))

        # 3.3 see 3.11

        # 3.4 multiply by window function
        windowed_frame = frame * window

        # 3.5-3.9
        lag_domain = _autocorr(windowed_frame, pad_width_for_pow2)

        # 3.10 divide by autocorrelation of window
        sampled_autocorr = lag_domain / autocorr_hann

        # only include up to half the window length because unreliable above (p. 100, fig)
        sampled_autocorr = sampled_autocorr[:(window_length_samples//2)]

        # 3.11 find places and heights of maxima
        unvoiced_strength = voicing_thresh + max(0, 2 - ((local_peak / global_peak) /
                                                         (silence_thresh / (1 + voicing_thresh))))

        voiced_candidates = _find_pitch_candidates_(
                sampled_autocorr,
                sr,
                min_pitch,
                max_pitch,
                max_candidates,
                octave_cost
            )

        candidates = [(0.0, unvoiced_strength)] + voiced_candidates
        all_candidates.append(candidates)
        intensities.append(local_peak / global_peak if global_peak > 0 else 0.0)

    # frame timing: match Yannicks code (centered on window)
    n_frames = len(framed_signal)
    t0 = 0.5 * (len(data) / sr - n_frames * timestep + timestep)
    time_points = t0 + np.arange(n_frames) * timestep

    # Praat-style: scale transition costs by time step as in Yannicks code
    dt = timestep
    time_step_correction = 0.01 / dt if dt else 1.0
    octave_jump_cost_scaled = octave_jump_cost * time_step_correction
    voiced_unvoiced_cost_scaled = voiced_unvoiced_cost * time_step_correction

    pitch_track = _viterbi_pitch_path(
        all_candidates,
        voiced_unvoiced_cost=voiced_unvoiced_cost_scaled,
        octave_jump_cost=octave_jump_cost_scaled
    )

    if plot:
        from biosonic.plot import plot_pitch_on_spectrogram
        plot_pitch_on_spectrogram(
            data,
            sr,
            time_points,
            pitch_track,
            **kwargs)

    return np.asarray(time_points), np.asarray(pitch_track), all_candidates, np.asarray(intensities)
